"""
The place to reacting information from the server
"""
import queue
import time
from typing import TYPE_CHECKING, List, Optional

from mcdreforged.constants import core_constant
from mcdreforged.info_reactor.abstract_info_reactor import AbstractInfoReactor
from mcdreforged.info_reactor.impl import PlayerReactor, ServerReactor, GeneralReactor
from mcdreforged.info_reactor.info import Info
from mcdreforged.utils import misc_util
from mcdreforged.utils.logger import ServerLogger, DebugOption

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer


class InfoReactorManager:
	def __init__(self, mcdr_server: 'MCDReforgedServer'):
		self.mcdr_server = mcdr_server
		self.last_queue_full_warn_time = None
		self.server_logger = ServerLogger('Server')
		self.reactors = []  # type: List[AbstractInfoReactor]

	def register_reactors(self, custom_reactor_class_paths: Optional[List[str]]):
		self.reactors.clear()
		self.reactors.extend([
			GeneralReactor(self.mcdr_server),
			ServerReactor(self.mcdr_server),
			PlayerReactor(self.mcdr_server)
		])
		if custom_reactor_class_paths is not None:
			for class_path in custom_reactor_class_paths:
				try:
					reactor_class = misc_util.load_class(class_path)
				except:
					self.mcdr_server.logger.exception('Fail to load info reactor from "{}"'.format(class_path))
				else:
					if issubclass(reactor_class, AbstractInfoReactor):
						self.reactors.append(reactor_class(self.mcdr_server))
						self.mcdr_server.logger.debug('Loaded info reactor {} from {}'.format(reactor_class.__name__, class_path), option=DebugOption.REACTOR)
					else:
						self.mcdr_server.logger.error('Wrong reactor class "{}", expected {} but found {}'.format(class_path, AbstractInfoReactor, reactor_class))

	def process_info(self, info: Info):
		for reactor in self.reactors:
			try:
				reactor.react(info)
			except:
				self.mcdr_server.logger.exception(self.mcdr_server.tr('info_reactor_manager.react.error', type(reactor).__name__))

		# send command input from the console to the server's stdin
		if info.is_from_console and info.should_send_to_server():
			self.mcdr_server.send(info.content)

	def put_info(self, info):
		info.attach_mcdr_server(self.mcdr_server)
		# echo info from the server to the console
		if info.is_from_server:
			self.server_logger.info(info.raw_content)
		try:
			self.mcdr_server.task_executor.enqueue_info_task(lambda: self.process_info(info), info.is_user)
		except queue.Full:
			current_time = time.monotonic()
			logging_method = self.mcdr_server.logger.debug
			kwargs = {'option': DebugOption.REACTOR}
			if self.last_queue_full_warn_time is None or current_time - self.last_queue_full_warn_time >= core_constant.REACTOR_QUEUE_FULL_WARN_INTERVAL_SEC:
				logging_method = self.mcdr_server.logger.warning
				kwargs = {}
				self.last_queue_full_warn_time = current_time
			logging_method(self.mcdr_server.tr('info_reactor_manager.info_queue.full'), **kwargs)

	def on_server_start(self):
		for reactor in self.reactors:
			reactor.on_server_start()

	def on_server_stop(self):
		for reactor in self.reactors:
			reactor.on_server_stop()



import threading
from typing import TYPE_CHECKING, Optional

from mcdreforged.utils import misc_util

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer


class ThreadExecutor:
	def __init__(self, mcdr_server: 'MCDReforgedServer'):
		self.mcdr_server = mcdr_server
		self._executor_thread = None  # type: Optional[threading.Thread]
		self.__stopped_looping = False
		self.__name = self.__class__.__name__

	def get_thread(self) -> Optional[threading.Thread]:
		return self._executor_thread

	def is_on_thread(self):
		return threading.current_thread() is self._executor_thread

	def should_keep_looping(self) -> bool:
		return not self.__stopped_looping

	def start(self):
		self.mcdr_server.logger.debug('{} is starting'.format(self.get_name()))
		self._executor_thread = misc_util.start_thread(self.loop, (), self.get_name())
		return self._executor_thread

	def stop(self):
		self.__stopped_looping = True

	def get_name(self):
		return self.__name

	def set_name(self, name: str):
		self.__name = name
		self._executor_thread.setName(name)

	def loop(self):
		while self.should_keep_looping():
			self.tick()

	def tick(self):
		raise NotImplementedError()

	def join(self):
		if self._executor_thread is None:
			raise RuntimeError()
		self._executor_thread.join()



import collections
import time
from typing import Dict, Optional, Tuple, List, TYPE_CHECKING, Set

from mcdreforged.handler.abstract_server_handler import AbstractServerHandler
from mcdreforged.handler.impl import *
from mcdreforged.utils import misc_util
from mcdreforged.utils.logger import DebugOption

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer


class ServerHandlerManager:
	HANDLER_DETECTION_MINIMUM_SAMPLE_COUNT = 20   # At least 20 messages
	HANDLER_DETECTION_MINIMUM_SAMPLING_TIME = 60  # will do sample for 1 minute

	def __init__(self, mcdr_server: 'MCDReforgedServer'):
		self.mcdr_server = mcdr_server
		self.logger = mcdr_server.logger
		self.handlers = {}  # type: Dict[str, AbstractServerHandler]
		self.__current_handler = None  # type: Optional[AbstractServerHandler]
		# the handler that should always work
		self.basic_handler = None  # type: Optional[AbstractServerHandler]

		# Automation for lazy
		self.__detection_running = False
		self.__detection_text_count = 0
		self.__detection_success_count = collections.defaultdict(lambda: 0)

	def register_handlers(self, custom_handler_class_paths: Optional[List[str]]):
		def add_handler(hdr: AbstractServerHandler):
			self.handlers[hdr.get_name()] = hdr

		self.handlers.clear()
		self.basic_handler = BasicHandler()
		add_handler(self.basic_handler)
		add_handler(VanillaHandler())
		add_handler(BukkitHandler())
		add_handler(Bukkit14Handler())
		add_handler(ForgeHandler())
		add_handler(CatServerHandler())
		add_handler(Beta18Handler())
		add_handler(BungeecordHandler())
		add_handler(WaterfallHandler())
		add_handler(VelocityHandler())
		if custom_handler_class_paths is not None:
			for class_path in custom_handler_class_paths:
				try:
					handler_class = misc_util.load_class(class_path)
				except:
					self.mcdr_server.logger.exception('Fail to load info handler from "{}"'.format(class_path))
				else:
					if issubclass(handler_class, AbstractServerHandler):
						handler = handler_class()
						if handler.get_name() not in self.handlers:
							add_handler(handler)
							self.mcdr_server.logger.debug('Loaded info handler {} from {}'.format(handler_class.__name__, class_path), option=DebugOption.HANDLER)
						else:
							self.mcdr_server.logger.error('Handler with name {} from path {} is already registered, ignored'.format(handler.get_name(), class_path))
					else:
						self.mcdr_server.logger.error('Wrong handler class "{}", expected {} but found {}'.format(class_path, AbstractServerHandler, handler_class))

	def set_handler(self, handler_name: str):
		try:
			self.__current_handler = self.handlers[handler_name]
		except KeyError:
			self.logger.error('Fail to load handler with name "{}"'.format(handler_name))
			self.logger.error('Fallback basic handler is used, MCDR might not works correctly'.format(handler_name))
			self.__current_handler = self.basic_handler

	def get_current_handler(self) -> AbstractServerHandler:
		return self.__current_handler

	def get_basic_handler(self) -> AbstractServerHandler:
		return self.basic_handler

	# Automation for lazy

	def start_handler_detection(self):
		if not self.is_detection_running():
			self.__detection_running = True
			self.__detection_text_count = 0
			self.__detection_success_count.clear()
			misc_util.start_thread(self.__detection_thread, (), 'HandlerDetector')

	def is_detection_running(self) -> bool:
		return self.__detection_running

	def __detection_thread(self):
		time.sleep(self.HANDLER_DETECTION_MINIMUM_SAMPLING_TIME)
		while self.__detection_text_count < self.HANDLER_DETECTION_MINIMUM_SAMPLE_COUNT:
			time.sleep(1)
		lst = self.finalize_detection_result()  # type: List[Tuple[AbstractServerHandler, int]]
		if len(lst) == 0:
			return
		total = self.__detection_text_count
		best_count = lst[0][1]
		end = 1
		while end < len(lst) and lst[end][1] == best_count:
			end += 1
		best_handlers = set(map(lambda item: item[0], lst[:end]))
		current_handler = self.get_current_handler()
		current_count = self.__detection_success_count[self.get_current_handler()]
		if current_handler not in best_handlers:
			self.mcdr_server.logger.warning(self.mcdr_server.tr('server_handler_manager.handler_detection.result1'))
			self.mcdr_server.logger.warning(self.mcdr_server.tr('server_handler_manager.handler_detection.result2', current_handler.get_name(), round(100.0 * current_count / total, 2), current_count, total))
			for best_handler, best_count in lst[:end]:
				self.mcdr_server.logger.warning(self.mcdr_server.tr('server_handler_manager.handler_detection.result3', best_handler.get_name(), round(100.0 * best_count / total, 2), best_count, total))

	def detect_text(self, text: str):
		if self.is_detection_running():
			self.__detection_text_count += 1
			for handler in self.handlers.values():
				if handler is not self.basic_handler:
					try:
						handler.parse_server_stdout(handler.pre_parse_server_stdout(text))
					except:
						pass
					else:
						self.__detection_success_count[handler] += 1

	def finalize_detection_result(self):
		self.__detection_running = False
		current_handler_set = set(self.handlers.values())  # type: Set[AbstractServerHandler]
		lst = list(filter(lambda item: item[0] in current_handler_set, self.__detection_success_count.items()))  # type: List[Tuple[AbstractServerHandler, int]]
		lst.sort(key=lambda item: item[1], reverse=True)
		return lst



import re

from mcdreforged.command.builder.exception import RequirementNotMet, UnknownArgument, CommandError
from mcdreforged.command.builder.nodes.basic import Literal
from mcdreforged.command.command_source import CommandSource
from mcdreforged.constants import core_constant
from mcdreforged.minecraft.rtext import RText, RAction, RTextList, RColor
from mcdreforged.permission.permission_level import PermissionLevel
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.check_update_command import CheckUpdateCommand
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.help_command import HelpCommand
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.permission_command import PermissionCommand
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.plugin_command import PluginCommand
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.preference_command import PreferenceCommand
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.reload_command import ReloadCommand
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.status_command import StatusCommand
from mcdreforged.plugin.meta.metadata import Metadata
from mcdreforged.plugin.type.permanent_plugin import PermanentPlugin
from mcdreforged.translation.translation_text import RTextMCDRTranslation

METADATA = {
	'id': core_constant.PACKAGE_NAME,
	'version': core_constant.VERSION,
	'name': core_constant.NAME,
	'description': 'The core of {}'.format(core_constant.NAME),
	'author': [
		'Fallen_Breath'
	],
	'link': 'https://github.com/Fallen-Breath/MCDReforged'
}


class MCDReforgedPlugin(PermanentPlugin):
	def __init__(self, plugin_manager):
		super().__init__(plugin_manager)
		self._set_metadata(Metadata(METADATA, plugin=self))
		self.command_help = HelpCommand(self)
		self.command_status = StatusCommand(self)
		self.command_reload = ReloadCommand(self)
		self.command_permission = PermissionCommand(self)
		self.command_plugin = PluginCommand(self)
		self.command_check_update = CheckUpdateCommand(self)
		self.command_preference = PreferenceCommand(self)

	def tr(self, key: str, *args, **kwargs) -> RTextMCDRTranslation:
		return self.server_interface.rtr(key, *args, **kwargs)

	def load(self):
		self.plugin_registry.clear()
		self.__register_commands()
		self.server_interface.register_help_message(self.control_command_prefix, self.tr('mcdr_command.help_message.mcdr_command'))
		self.server_interface.register_help_message(self.help_command_prefix, self.tr('mcdr_command.help_message.help_command'))

	def __repr__(self):
		# avoid using self.metadata here since it might not be initialized
		return 'MCDReforgedPlugin[version={}]'.format(METADATA['version'])

	@property
	def control_command_prefix(self):
		return '!!MCDR'

	@property
	def help_command_prefix(self):
		return '!!help'

	def __register_commands(self):
		self.register_command(
			Literal(self.control_command_prefix).
			requires(lambda src: src.has_permission(PermissionLevel.USER)).
			runs(self.process_mcdr_command).
			on_error(RequirementNotMet, self.on_mcdr_command_permission_denied, handled=True).
			on_child_error(RequirementNotMet, self.on_mcdr_command_permission_denied, handled=True).
			on_error(UnknownArgument, self.on_mcdr_command_unknown_argument, handled=True).
			then(self.command_status.get_command_node()).
			then(self.command_reload.get_command_node()).
			then(self.command_permission.get_command_node()).
			then(self.command_plugin.get_command_node()).
			then(self.command_check_update.get_command_node()).
			then(self.command_preference.get_command_node())
		)
		self.register_command(self.command_help.get_command_node())

	# --------------------
	#    Command Stuffs
	# --------------------

	def get_help_message(self, source: CommandSource, translation_key: str):
		lst = RTextList()
		with source.preferred_language_context():
			for line in self.tr(translation_key).to_plain_text().splitlines(keepends=True):
				prefix = re.search(r'(?<=§7)' + self.control_command_prefix + r'[\w ]*(?=§)', line)
				if prefix is not None:
					lst.append(RText(line).c(RAction.suggest_command, prefix.group()))
				else:
					lst.append(line)
		return lst

	def on_mcdr_command_permission_denied(self, source: CommandSource, error: CommandError):
		if not error.is_handled():
			source.reply(RText(self.mcdr_server.tr('mcdr_command.permission_denied'), color=RColor.red))

	def on_mcdr_command_unknown_argument(self, source: CommandSource, error: CommandError):
		if source.has_permission(PermissionLevel.MCDR_CONTROL_LEVEL):
			command = error.get_parsed_command().rstrip(' ')
			source.reply(
				RText(self.tr('mcdr_command.command_not_found', command)).
				h(self.tr('mcdr_command.command_not_found_suggest', command)).
				c(RAction.run_command, command)
			)
		else:
			self.on_mcdr_command_permission_denied(source, error)

	def process_mcdr_command(self, source: CommandSource):
		if source.has_permission(PermissionLevel.MCDR_CONTROL_LEVEL):
			source.reply(self.get_help_message(source, 'mcdr_command.help_message'))
		else:
			self.command_status.print_mcdr_status(source)



"""
Some customize exceptions
"""


# Fail to start the server
class ServerStartError(RuntimeError):
	pass


# The function call is illegal due to some reasons
class IllegalCallError(RuntimeError):
	pass


class IllegalStateError(RuntimeError):
	pass


class ServerStopped(RuntimeError):
	pass


class IllegalPluginStructure(Exception):
	pass


# When the metadata of a plugin is invalid
class BrokenMetadata(Exception):
	pass



from typing import Union, Any

DIVIDER = ' '


def remove_divider_prefix(text: str) -> str:
	return text.lstrip(DIVIDER)


def get_element(text) -> str:
	"""
	"my test command" -> "my"
	:type text: str
	:rtype: str
	"""
	pos = text.find(DIVIDER)
	if pos == -1:
		return text
	else:
		return text[:pos]


def __get_var(text, func) -> Union[Any, int]:
	"""
	Return value, char_read
	value will be None if ValueError occurs
	"""
	arg = get_element(text)
	try:
		value = func(arg)
	except ValueError:
		value = None
	return value, len(arg)


def get_int(text) -> Union[int, int]:
	return __get_var(text, int)


def get_float(text) -> Union[float, int]:
	return __get_var(text, float)



"""
Analyzing and reacting events related to player
"""
from mcdreforged.info_reactor.abstract_info_reactor import AbstractInfoReactor
from mcdreforged.info_reactor.info import InfoSource
from mcdreforged.plugin.plugin_event import MCDRPluginEvents
from mcdreforged.utils.logger import DebugOption


class PlayerReactor(AbstractInfoReactor):
	def react(self, info):
		if info.source == InfoSource.SERVER:
			handler = self.mcdr_server.server_handler_manager.get_current_handler()

			# on_player_joined
			player = handler.parse_player_joined(info)
			if player is not None:
				self.mcdr_server.logger.debug('Player joined detected', option=DebugOption.REACTOR)
				self.mcdr_server.permission_manager.touch_player(player)
				self.mcdr_server.plugin_manager.dispatch_event(MCDRPluginEvents.PLAYER_JOINED, (player, info))

			# on_player_left
			player = handler.parse_player_left(info)
			if player is not None:
				self.mcdr_server.logger.debug('Player left detected', option=DebugOption.REACTOR)
				self.mcdr_server.plugin_manager.dispatch_event(MCDRPluginEvents.PLAYER_LEFT, (player,))

			# # on_death_message
			# if handler.parse_death_message(info):
			# 	self.mcdr_server.logger.debug('Death message detected')
			# 	self.mcdr_server.plugin_manager.call('on_death_message', (self.mcdr_server.server_interface, info.content))
			#
			# # on_player_made_advancement
			# result = handler.parse_player_made_advancement(info)
			# if result is not None:
			# 	self.mcdr_server.logger.debug('Player made advancement detected')
			# 	player, advancement = result
			# 	self.mcdr_server.plugin_manager.call('on_player_made_advancement', (self.mcdr_server.server_interface, player, advancement))



"""
Custom logger for MCDR
"""
import logging
import os
import sys
import time
import weakref
import zipfile
from contextlib import contextmanager
from enum import Enum, unique, auto
from threading import RLock, local
from typing import Dict, Optional, List, Set

from colorlog import ColoredFormatter

from mcdreforged.constants import core_constant
from mcdreforged.minecraft.rtext import RColor, RStyle, RItem
from mcdreforged.utils import string_util, file_util


@unique
class DebugOption(Enum):
	# remember to sync with the debug field in the default config file
	ALL = auto()
	MCDR = auto()
	HANDLER = auto()
	REACTOR = auto()
	PLUGIN = auto()
	PERMISSION = auto()
	COMMAND = auto()
	TASK_EXECUTOR = auto()


class SyncStdoutStreamHandler(logging.StreamHandler):
	__write_lock = RLock()
	__instance_lock = RLock()
	__instances = weakref.WeakSet()  # type: Set[SyncStdoutStreamHandler]

	def __init__(self):
		super().__init__(sys.stdout)
		with self.__instance_lock:
			self.__instances.add(self)

	def emit(self, record) -> None:
		with self.__write_lock:
			super().emit(record)

	@classmethod
	def update_stdout(cls, stream=None):
		if stream is None:
			stream = sys.stdout
		with cls.__instance_lock:
			instances = list(cls.__instances)
		for inst in instances:
			inst.acquire()  # use Handler's lock
			try:
				inst.stream = stream
			finally:
				inst.release()


class MCColoredFormatter(ColoredFormatter):
	MC_CODE_ITEMS = list(map(lambda item: item.value, list(RColor) + list(RStyle)))  # type: List[RItem]

	# global flag
	console_color_disabled = False

	__TLS = local()

	@classmethod
	@contextmanager
	def disable_minecraft_color_code_transform(cls):
		cls.__set_mc_code_trans_disable(True)
		try:
			yield
		finally:
			cls.__set_mc_code_trans_disable(False)

	@classmethod
	def __is_mc_code_trans_disabled(cls) -> bool:
		try:
			return cls.__TLS.mc_code_trans
		except AttributeError:
			cls.__set_mc_code_trans_disable(False)
			return False

	@classmethod
	def __set_mc_code_trans_disable(cls, state: bool):
		cls.__TLS.mc_code_trans = state

	def formatMessage(self, record):
		text = super().formatMessage(record)
		if not self.__is_mc_code_trans_disabled():
			# minecraft code -> console code
			for item in self.MC_CODE_ITEMS:
				if item.mc_code in text:
					text = text.replace(item.mc_code, item.console_code)
			# clean the rest of minecraft codes
			text = string_util.clean_minecraft_color_code(text)
		if self.console_color_disabled:
			text = string_util.clean_console_color_code(text)
		return text


class NoColorFormatter(logging.Formatter):
	def formatMessage(self, record):
		return string_util.clean_console_color_code(super().formatMessage(record))


class MCDReforgedLogger(logging.Logger):
	DEFAULT_NAME = core_constant.NAME_SHORT
	LOG_COLORS = {
		'DEBUG': 'blue',
		'INFO': 'green',
		'WARNING': 'yellow',
		'ERROR': 'red',
		'CRITICAL': 'bold_red',
	}
	SECONDARY_LOG_COLORS = {
		'message': {
			'WARNING': 'yellow',
			'ERROR': 'red',
			'CRITICAL': 'red'
		}
	}
	FILE_FMT = NoColorFormatter(
		'[%(name)s] [%(asctime)s] [%(threadName)s/%(levelname)s]: %(message)s',
		datefmt='%Y-%m-%d %H:%M:%S'
	)

	@classmethod
	def get_console_formatter(cls, plugin_id=None):
		extra = '' if plugin_id is None else ' [{}]'.format(plugin_id)
		return MCColoredFormatter(
			f'[%(name)s] [%(asctime)s] [%(threadName)s/%(log_color)s%(levelname)s%(reset)s]{extra}: %(message_log_color)s%(message)s%(reset)s',
			log_colors=cls.LOG_COLORS,
			secondary_log_colors=cls.SECONDARY_LOG_COLORS,
			datefmt='%H:%M:%S'
		)

	debug_options = {}  # type: Dict[DebugOption, bool]

	def __init__(self, plugin_id: Optional[str] = None):
		super().__init__(self.DEFAULT_NAME)
		self.file_handler = None

		self.console_handler = SyncStdoutStreamHandler()
		self.console_handler.setFormatter(self.get_console_formatter(plugin_id))

		self.addHandler(self.console_handler)
		self.setLevel(logging.DEBUG)

	def set_debug_options(self, debug_options: Dict[str, bool]):
		for key, value in debug_options.items():
			option = DebugOption[key.upper()]
			MCDReforgedLogger.debug_options[option] = value
		for option, value in MCDReforgedLogger.debug_options.items():
			self.debug('Debug option {} is set to {}'.format(option, value), option=option)

	@classmethod
	def should_log_debug(cls, option: Optional[DebugOption] = None):
		do_log = cls.debug_options.get(DebugOption.ALL, False)
		if option is not None:
			do_log |= cls.debug_options.get(option, False)
		return do_log

	def debug(self, *args, option: Optional[DebugOption] = None, no_check: bool = False):
		if no_check or self.should_log_debug(option):
			with MCColoredFormatter.disable_minecraft_color_code_transform():
				super().debug(*args)

	def set_file(self, file_name: str):
		if self.file_handler is not None:
			self.removeHandler(self.file_handler)
		file_util.touch_directory(os.path.dirname(file_name))
		if os.path.isfile(file_name):
			modify_time = time.strftime('%Y-%m-%d', time.localtime(os.stat(file_name).st_mtime))
			counter = 0
			while True:
				counter += 1
				zip_file_name = '{}/{}-{}.zip'.format(os.path.dirname(file_name), modify_time, counter)
				if not os.path.isfile(zip_file_name):
					break
			zipf = zipfile.ZipFile(zip_file_name, 'w')
			zipf.write(file_name, arcname=os.path.basename(file_name), compress_type=zipfile.ZIP_DEFLATED)
			zipf.close()
			os.remove(file_name)
		self.file_handler = logging.FileHandler(file_name, encoding='utf8')
		self.file_handler.setFormatter(self.FILE_FMT)
		self.addHandler(self.file_handler)


class ServerLogger(logging.Logger):
	server_fmt = MCColoredFormatter('[%(name)s] %(message)s')

	def __init__(self, name):
		super().__init__(name)

		console_handler = SyncStdoutStreamHandler()
		console_handler.setFormatter(self.server_fmt)

		self.addHandler(console_handler)
		self.setLevel(logging.DEBUG)



from typing import Optional

from mcdreforged.constants import core_constant
from mcdreforged.utils.types import TranslationKeyDictRich, MessageText, TranslationStorage, TranslationKeyDictNested

__all__ = [
	'translate_from_dict',
	'update_storage',
]

_NONE = object()


def translate_from_dict(translations: TranslationKeyDictRich, language: str, *, fallback_language: Optional[str] = core_constant.DEFAULT_LANGUAGE, default: Optional[MessageText] = _NONE) -> MessageText:
	"""
	Select a translation for given language based on a translation dict
	:param language: The language
	:param fallback_language: If translation file, it will try using the fallback_language as language for a second try
	:param translations: A language -> text mapping
	:param default: The fallback value. If not specified and translation not found, KeyError will be risen
	"""
	result = translations.get(language)
	if result is None and fallback_language is not None:
		result = translations.get(fallback_language)
	if result is None:
		result = default
	if result is _NONE:
		raise KeyError('Failed to translate from dict with translations {}, language {}, fallback_language {}'.format(translations, language, fallback_language))
	return result


def update_storage(storage: TranslationStorage, language: str, mapping: TranslationKeyDictNested):
	__update_storage(storage, language, mapping)


def __update_storage(storage: TranslationStorage, language: str, mapping: TranslationKeyDictNested, path: Optional[str] = None):
	for key, item in mapping.items():
		current_path = f'{path}.{key}' if path is not None else key
		if isinstance(item, str):
			storage[current_path][language] = item
		elif isinstance(item, dict):
			__update_storage(storage, language, item, current_path)



import threading
from contextlib import contextmanager
from typing import Union, Iterable, Optional, List, Callable

from mcdreforged.constants import core_constant
from mcdreforged.minecraft.rtext import RTextBase, RAction, RStyle, RColor, RText
from mcdreforged.utils import translation_util


class RTextMCDRTranslation(RTextBase):
	__TLS = threading.local()
	__TLS.language = None
	__TLS.fallback_language = None

	def __init__(self, translation_key: str, *args, **kwargs):
		self.translation_key: str = translation_key
		self.args = args
		self.kwargs = kwargs
		self.__translator = lambda *args_, **kwargs_: RText(self.translation_key)
		self.__post_process: List[Callable[[RTextBase], RTextBase]] = []

		from mcdreforged.plugin.server_interface import ServerInterface
		server: Optional[ServerInterface] = ServerInterface.get_instance()
		if server is not None:
			self.set_translator(server.tr)

	def set_translator(self, translate_function: Callable) -> 'RTextMCDRTranslation':
		self.__translator = translate_function
		return self

	@classmethod
	def from_translation_dict(cls, translation_dict) -> 'RTextMCDRTranslation':
		def fake_tr(key: str, language: str):
			return translation_util.translate_from_dict(translation_dict, language)

		return RTextMCDRTranslation('').set_translator(fake_tr)

	def __get_translated_text(self) -> RTextBase:
		language = getattr(self.__TLS, 'language', None)
		if language is None:
			from mcdreforged.plugin.server_interface import ServerInterface
			server: Optional[ServerInterface] = ServerInterface.get_instance()
			if server is not None:
				language = server.get_mcdr_language()
			else:
				language = core_constant.DEFAULT_LANGUAGE
		processed_text = self.__translator(self.translation_key, *self.args, **self.kwargs, language=language)
		processed_text = RTextBase.from_any(processed_text)
		for process in self.__post_process:
			processed_text = process(processed_text)
		return processed_text

	@classmethod
	@contextmanager
	def language_context(cls, language: str):
		prev = getattr(cls.__TLS, 'language', None)
		cls.__TLS.language = language
		try:
			yield
		finally:
			cls.__TLS.language = prev

	def to_json_object(self) -> Union[dict, list]:
		return self.__get_translated_text().to_json_object()

	def to_plain_text(self) -> str:
		return self.__get_translated_text().to_plain_text()

	def to_colored_text(self) -> str:
		return self.__get_translated_text().to_colored_text()

	def copy(self) -> 'RTextBase':
		copied = RTextMCDRTranslation(self.translation_key, *self.args, **self.kwargs)
		copied.set_translator(self.__translator)
		return copied

	def set_color(self, color: RColor) -> 'RTextBase':
		self.__post_process.append(lambda rt: rt.set_color(color))
		return self

	def set_styles(self, styles: Union[RStyle, Iterable[RStyle]]) -> 'RTextBase':
		self.__post_process.append(lambda rt: rt.set_styles(styles))
		return self

	def set_click_event(self, action: RAction, value: str) -> 'RTextBase':
		self.__post_process.append(lambda rt: rt.set_click_event(action, value))
		return self

	def set_hover_text(self, *args) -> 'RTextBase':
		self.__post_process.append(lambda rt: rt.set_hover_text(*args))
		return self



from mcdreforged.command.builder.exception import UnknownArgument
from mcdreforged.command.builder.nodes.basic import Literal
from mcdreforged.command.command_source import CommandSource
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.sub_command import SubCommand


class ReloadCommand(SubCommand):
	def get_command_node(self) -> Literal:
		return (
			self.control_command_root({'r', 'reload'}).
			runs(lambda src: src.reply(self.get_help_message(src, 'mcdr_command.help_message.reload'))).
			on_error(UnknownArgument, self.on_mcdr_command_unknown_argument).
			then(Literal({'plugin', 'plg'}).runs(self.refresh_changed_plugins)).
			then(Literal({'config', 'cfg'}).runs(self.reload_config)).
			then(Literal({'permission', 'perm'}).runs(self.reload_permission)).
			then(Literal('all').runs(self.reload_all))
		)

	def refresh_changed_plugins(self, source: CommandSource):
		ret = self.function_call(source, self.mcdr_server.plugin_manager.refresh_changed_plugins, 'refresh_changed_plugins', log_success=False)
		self._print_plugin_operation_result_if_no_error(source, ret)

	def reload_config(self, source: CommandSource):
		self.function_call(source, self.mcdr_server.load_config, 'reload_config')

	def reload_permission(self, source: CommandSource):
		self.function_call(source, self.mcdr_server.permission_manager.load_permission_file, 'reload_permission')

	def reload_all(self, source: CommandSource):
		self.reload_config(source)
		self.reload_permission(source)
		self.refresh_changed_plugins(source)



"""
Advanced text container for Minecraft
Credit: Pandaria98 https://github.com/Pandaria98 https://github.com/TISUnion/stext
"""

import json
from enum import Enum, auto
from typing import Iterable, List, Union, Optional, Any, Tuple, Set, NamedTuple

from colorama import Fore, Style


class RItem:
	def __init__(self, mc_code: str, console_code: str):
		self.mc_code = mc_code
		self.console_code = console_code


class RColor(Enum):
	black = RItem('§0', Fore.BLACK)
	dark_blue = RItem('§1', Fore.BLUE)
	dark_green = RItem('§2', Fore.GREEN)
	dark_aqua = RItem('§3', Fore.CYAN)
	dark_red = RItem('§4', Fore.RED)
	dark_purple = RItem('§5', Fore.MAGENTA)
	gold = RItem('§6', Fore.YELLOW)
	gray = RItem('§7', Style.RESET_ALL)
	dark_gray = RItem('§8', Style.RESET_ALL)
	blue = RItem('§9', Fore.LIGHTBLUE_EX)
	green = RItem('§a', Fore.LIGHTGREEN_EX)
	aqua = RItem('§b', Fore.LIGHTCYAN_EX)
	red = RItem('§c', Fore.LIGHTRED_EX)
	light_purple = RItem('§d', Fore.LIGHTMAGENTA_EX)
	yellow = RItem('§e', Fore.LIGHTYELLOW_EX)
	white = RItem('§f', Style.RESET_ALL)

	reset = RItem('§r', Style.RESET_ALL)


class RColorConvertor:
	MC_COLOR_TO_CONSOLE = dict([(rcolor.value.mc_code, rcolor.value.console_code) for rcolor in RColor])
	MC_COLOR_TO_RCOLOR = dict([(rcolor.value.mc_code, rcolor) for rcolor in RColor])
	RCOLOR_TO_CONSOLE = dict([(rcolor, rcolor.value.console_code) for rcolor in RColor])
	RCOLOR_NAME_TO_CONSOLE = dict([(rcolor.name, rcolor.value.console_code) for rcolor in RColor])


class RStyle(Enum):
	bold = RItem('§l', Style.BRIGHT)
	italic = RItem('§o', '')
	underlined = RItem('§n', '')
	strikethrough = RItem('§m', '')
	obfuscated = RItem('§k', '')


class RAction(Enum):
	suggest_command = auto()
	run_command = auto()
	open_url = auto()
	open_file = auto()
	copy_to_clipboard = auto()


class RTextBase:
	def to_json_object(self) -> Union[dict, list]:
		raise NotImplementedError()

	def to_json_str(self) -> str:
		return json.dumps(self.to_json_object(), ensure_ascii=False)

	def to_plain_text(self) -> str:
		raise NotImplementedError()

	def to_colored_text(self) -> str:
		raise NotImplementedError()

	def copy(self) -> 'RTextBase':
		raise NotImplementedError()

	def set_color(self, color: RColor) -> 'RTextBase':
		raise NotImplementedError()

	def set_styles(self, styles: Union[RStyle, Iterable[RStyle]]) -> 'RTextBase':
		raise NotImplementedError()

	def set_click_event(self, action: RAction, value: str) -> 'RTextBase':
		raise NotImplementedError()

	def set_hover_text(self, *args) -> 'RTextBase':
		raise NotImplementedError()

	def c(self, action: RAction, value: str) -> 'RTextBase':
		return self.set_click_event(action, value)

	def h(self, *args) -> 'RTextBase':
		return self.set_hover_text(*args)

	def __str__(self):
		return self.to_plain_text()

	def __add__(self, other):
		return RTextList(self, other)

	def __radd__(self, other):
		return RTextList(other, self)

	@staticmethod
	def from_any(text) -> 'RTextBase':
		"""
		:param text: can be a RTextBase, or a str, or anything
		:rtype: RTextBase
		"""
		if isinstance(text, RTextBase):
			return text
		return RText(text)

	@staticmethod
	def join(divider: Any, iterable: Iterable[Any]) -> 'RTextBase':
		result = RTextList()
		for i, item in enumerate(iterable):
			if i > 0:
				result.append(divider)
			result.append(item)
		return result

	@staticmethod
	def format(fmt: str, *args, **kwargs) -> 'RTextBase':
		args = list(args)
		kwargs = kwargs.copy()
		counter = 0
		rtext_elements = []  # type: List[Tuple[str, RTextBase]]

		def get():
			nonlocal counter
			rv = '@@MCDR#RText.format#Placeholder#{}@@'.format(counter)
			counter += 1
			return rv

		for i, arg in enumerate(args):
			if isinstance(arg, RTextBase):
				placeholder = get()
				rtext_elements.append((placeholder, arg))
				args[i] = placeholder
		for key, value in kwargs.items():
			if isinstance(value, RTextBase):
				placeholder = get()
				rtext_elements.append((placeholder, value))
				kwargs[key] = placeholder

		texts = [fmt.format(*args, **kwargs)]
		for placeholder, rtext in rtext_elements:
			new_texts = []
			for text in texts:
				processed_text = []
				if isinstance(text, str):
					for j, ele in enumerate(text.split(placeholder)):
						if j > 0:
							processed_text.append(rtext)
						processed_text.append(ele)
				else:
					processed_text.append(text)
				new_texts.extend(processed_text)
			texts = new_texts
		return RTextList(*texts)


class _ClickEvent(NamedTuple):
	action: RAction
	value: str


class RText(RTextBase):
	def __init__(self, text, color: Optional[RColor] = None, styles: Optional[Union[RStyle, Iterable[RStyle]]] = None):
		self.__text: str = str(text)
		self.__color: Optional[RColor] = None
		self.__styles: Set[RStyle] = set()
		self.__click_event: Optional[_ClickEvent] = None
		self.__hover_text_list: list = []
		if color is not None:
			self.set_color(color)
		if styles is not None:
			self.set_styles(styles)

	def _copy_from(self, text: 'RText'):
		self.__text = text.__text
		self.__color = text.__color
		self.__styles = text.__styles.copy()
		self.__click_event = text.__click_event
		self.__hover_text_list = text.__hover_text_list.copy()

	def set_color(self, color: RColor):
		self.__color = color
		return self

	def set_styles(self, styles: Union[RStyle, Iterable[RStyle]]):
		if isinstance(styles, RStyle):
			styles = {styles}
		elif isinstance(styles, Iterable):
			styles = set(styles)
		else:
			raise TypeError('Unsupported style type {}'.format(type(styles)))
		self.__styles = styles
		return self

	def set_click_event(self, action: RAction, value: str):
		self.__click_event = _ClickEvent(action, value)
		return self

	def set_hover_text(self, *args):
		self.__hover_text_list = list(args)
		return self

	def to_json_object(self) -> Union[dict, list]:
		obj = {'text': self.__text}
		if self.__color is not None:
			obj['color'] = self.__color.name
		for style in self.__styles:
			obj[style.name] = True
		if self.__click_event is not None:
			obj['clickEvent'] = {
				'action': self.__click_event.action.name,
				'value': self.__click_event.value
			}
		if len(self.__hover_text_list) > 0:
			obj['hoverEvent'] = {
				'action': 'show_text',
				'value': {
					'text': '',
					'extra': RTextList(*self.__hover_text_list).to_json_object(),
				}
			}
		return obj

	def to_plain_text(self) -> str:
		return self.__text

	def to_colored_text(self) -> str:
		color = RColorConvertor.RCOLOR_NAME_TO_CONSOLE[self.__color.name] if self.__color is not None else ''
		for style in self.__styles:
			color += style.value.console_code
		return color + self.to_plain_text() + Style.RESET_ALL

	def copy(self) -> 'RText':
		copied = RText('')
		copied._copy_from(self)
		return copied


class RTextTranslation(RText):
	def __init__(self, translation_key: str, color: RColor = RColor.reset, styles: Optional[Union[RStyle, Iterable[RStyle]]] = None):
		super().__init__(translation_key, color, styles)
		self.__translation_key: str = translation_key

	def to_plain_text(self) -> str:
		return self.__translation_key

	def to_json_object(self) -> Union[dict, list]:
		obj = super().to_json_object()
		obj.pop('text')
		obj['translate'] = self.__translation_key
		return obj

	def _copy_from(self, text: 'RTextTranslation'):
		super()._copy_from(text)
		self.__translation_key = text.__translation_key

	def copy(self) -> 'RTextTranslation':
		copied = RTextTranslation('')
		copied._copy_from(self)
		return copied


class RTextList(RTextBase):
	def __init__(self, *args):
		self.header = RText('')
		self.header_empty = True
		self.children = []  # type: List[RTextBase]
		self.append(*args)

	def set_color(self, color: RColor):
		self.header.set_color(color)
		self.header_empty = False
		return self

	def set_styles(self, styles: Union[RStyle, Iterable[RStyle]]):
		self.header.set_styles(styles)
		self.header_empty = False
		return self

	def set_click_event(self, action: RAction, value):
		self.header.set_click_event(action, value)
		self.header_empty = False
		return self

	def set_hover_text(self, *args):
		self.header.set_hover_text(*args)
		self.header_empty = False
		return self

	def append(self, *args) -> 'RTextList':
		for obj in args:
			if isinstance(obj, RTextList):
				self.children.extend(obj.children)
			elif isinstance(obj, RTextBase):
				self.children.append(obj)
			else:
				self.children.append(RText(obj))
		return self

	def is_empty(self) -> bool:
		return len(self.children) == 0

	def to_json_object(self) -> Union[dict, list]:
		ret = ['' if self.header_empty else self.header.to_json_object()]
		ret.extend(map(lambda rtext: rtext.to_json_object(), self.children))
		return ret

	def to_plain_text(self) -> str:
		return ''.join(map(lambda rtext: rtext.to_plain_text(), self.children))

	def to_colored_text(self) -> str:
		return ''.join(map(lambda rtext: rtext.to_colored_text(), self.children))

	def copy(self) -> 'RTextList':
		copied = RTextList()
		copied.header = self.header.copy()
		copied.header_empty = self.header_empty
		copied.children = [child.copy() for child in self.children]
		return copied



"""
A more flexible interface for rcon support
It also wrap everything to make sure no exception can escape
"""
from typing import TYPE_CHECKING, Optional

from mcdreforged.minecraft.rcon.rcon_connection import RconConnection

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer


class RconManager:
	def __init__(self, mcdr_server: 'MCDReforgedServer'):
		self.mcdr_server = mcdr_server
		self.logger = mcdr_server.logger
		self.rcon = None  # type: Optional[RconConnection]

	def is_running(self) -> bool:
		return self.rcon is not None and self.rcon.socket is not None

	def connect(self, address, port, password):
		if self.is_running():
			self.disconnect()
		self.rcon = RconConnection(address, port, str(password), logger=self.logger)
		try:
			success = self.rcon.connect()
		except Exception as e:
			self.logger.exception(self.mcdr_server.tr('rcon_manager.connect.connection_fail', e))
			self.rcon = None
		else:
			if success:
				self.logger.info(self.mcdr_server.tr('rcon_manager.connect.connected'))
			else:
				self.logger.info(self.mcdr_server.tr('rcon_manager.connect.wrong_password'))

	def disconnect(self):
		if self.is_running():
			try:
				self.rcon.disconnect()
				self.logger.info(self.mcdr_server.tr('rcon_manager.disconnect.disconnected'))
			except:
				self.mcdr_server.logger.error(self.mcdr_server.tr('rcon_manager.disconnect.disconnect_fail'))
		self.rcon = None

	def send_command(self, command):
		if self.is_running():
			return self.rcon.send_command(command)
		else:
			return None



from typing import NamedTuple, Any, List

from mcdreforged.command.builder.nodes.arguments import Integer
from mcdreforged.command.builder.nodes.basic import Literal, CommandContext
from mcdreforged.command.command_source import CommandSource
from mcdreforged.minecraft.rtext import RText, RAction, RColor
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.sub_command import SubCommand
from mcdreforged.plugin.plugin_registry import HelpMessage

HELP_MESSAGE_PER_PAGE = 10


class FunctionCallResult(NamedTuple):
	return_value: Any
	no_error: bool


class HelpCommand(SubCommand):
	def get_command_node(self) -> Literal:
		return (
			Literal(self.help_command_prefix).
			runs(self.process_help_command).
			then(
				Integer('page').at_min(1).
				runs(self.process_help_command)
			)
		)

	def process_help_command(self, source: CommandSource, context: CommandContext):
		page = context.get('page')
		source.reply(self.tr('mcdr_command.help_message.title'))
		matched: List[HelpMessage] = []
		msg: HelpMessage
		for msg in self.mcdr_server.plugin_manager.registry_storage.help_messages:
			if source.has_permission(msg.permission):
				matched.append(msg)
		matched_count = len(matched)

		if page is not None:
			left, right = (page - 1) * HELP_MESSAGE_PER_PAGE, page * HELP_MESSAGE_PER_PAGE
		else:
			left, right = 0, matched_count
		for i in range(left, right):
			if 0 <= i < matched_count:
				msg = matched[i]
				source.reply(RText.format('{}: {}', RText(msg.prefix, color=RColor.gray).c(RAction.suggest_command, msg.prefix), msg.message))

		if page is not None:
			has_prev = 0 < left < matched_count
			has_next = 0 < right < matched_count
			color = {False: RColor.dark_gray, True: RColor.gray}
			prev_page = RText('<-', color=color[has_prev])
			if has_prev:
				prev_page.c(RAction.run_command, '{} {}'.format(self.help_command_prefix, page - 1)).h(self.tr('mcdr_command.help_message.previous_page_hover'))
			next_page = RText('->', color=color[has_next])
			if has_next:
				next_page.c(RAction.run_command, '{} {}'.format(self.help_command_prefix, page + 1)).h(self.tr('mcdr_command.help_message.next_page_hover'))

			source.reply(RText.format('{} {} {}', prev_page, self.tr('mcdr_command.help_message.page_number', page), next_page))



"""
For lazy man
"""

# noinspection PyUnresolvedReferences
from mcdreforged.api.command import *
# noinspection PyUnresolvedReferences
from mcdreforged.api.decorator import *
# noinspection PyUnresolvedReferences
from mcdreforged.api.event import *
# noinspection PyUnresolvedReferences
from mcdreforged.api.exception import *
# noinspection PyUnresolvedReferences
from mcdreforged.api.rcon import *
# noinspection PyUnresolvedReferences
from mcdreforged.api.rtext import *
# noinspection PyUnresolvedReferences
from mcdreforged.api.types import *
# noinspection PyUnresolvedReferences
from mcdreforged.api.utils import *



import os
from typing import Optional, TYPE_CHECKING

from mcdreforged.constants import plugin_constant
from mcdreforged.plugin.type.directory_plugin import DirectoryPlugin
from mcdreforged.plugin.type.packed_plugin import PackedPlugin
from mcdreforged.plugin.type.regular_plugin import RegularPlugin
from mcdreforged.plugin.type.solo_plugin import SoloPlugin
from mcdreforged.utils import string_util, file_util

if TYPE_CHECKING:
	from mcdreforged.plugin.plugin_manager import PluginManager


def __get_plugin_class_from_path(file_path: str, allow_disabled: bool) -> Optional[type]:
	if os.path.isfile(file_path):
		if file_path.endswith(plugin_constant.DISABLED_PLUGIN_FILE_SUFFIX) and allow_disabled:
			file_path = string_util.remove_suffix(file_path, plugin_constant.DISABLED_PLUGIN_FILE_SUFFIX)
		suffix = file_util.get_file_suffix(file_path)
		if suffix == plugin_constant.SOLO_PLUGIN_FILE_SUFFIX:  # .py
			return SoloPlugin
		if suffix in plugin_constant.PACKED_PLUGIN_FILE_SUFFIXES:  # .mcdr
			return PackedPlugin
	elif os.path.isdir(file_path) and (allow_disabled or not file_path.endswith(plugin_constant.DISABLED_PLUGIN_FILE_SUFFIX)):
		if os.path.isfile(os.path.join(file_path, plugin_constant.PLUGIN_META_FILE)) and not os.path.isfile(os.path.join(file_path, '__init__.py')):
			return DirectoryPlugin
	return None


def __maybe_plugin(file_path: str, *, allow_disabled: bool) -> bool:
	return __get_plugin_class_from_path(file_path, allow_disabled) is not None


def is_plugin(file_path: str) -> bool:
	return __maybe_plugin(file_path, allow_disabled=False)


def is_disabled_plugin(file_path: str) -> bool:
	return not __maybe_plugin(file_path, allow_disabled=False) and __maybe_plugin(file_path, allow_disabled=True)


def create_regular_plugin(plugin_manager: 'PluginManager', file_path: str) -> RegularPlugin:
	cls = __get_plugin_class_from_path(file_path, False)
	if cls is None:
		raise TypeError('Trying to create a regular plugin with invalid path {}'.format(file_path))
	return cls(plugin_manager, file_path)



"""
Resource files / Inner modules getters
Don't run it directly
"""
import pkgutil
from typing import Optional

__all__ = [
	'ROOT_PACKAGE',
	'get_data', 'get_yaml'
]

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap

from mcdreforged.constants import core_constant

ROOT_PACKAGE = core_constant.PACKAGE_NAME


def __get_path(path: str) -> str:
	if path.startswith('/'):
		path = path[1:]
	return path


def get_data(path: str) -> Optional[bytes]:
	return pkgutil.get_data(ROOT_PACKAGE, __get_path(path))


def get_yaml(path: str) -> CommentedMap:
	bytes_data = get_data(path)
	# Replace CRLF or yaml loader will load extra lines
	string_data = bytes_data.decode('utf8').replace('\r\n', '\n')
	ret = YAML().load(string_data)
	return ret


if __name__ == '__main__':
	# Don't run it directly
	pass



from threading import Thread, current_thread, Lock
from typing import Dict, Optional, Any


class ThreadLocalStorage:
	def __init__(self):
		self.__storage = {}  # type: Dict[Thread, Dict]
		self.__lock = Lock()

	def __get_dict(self, thread: Thread):
		if thread not in self.__storage:
			self.__storage[thread] = {}
		return self.__storage[thread]

	def get(self, key: Any, default=None, *, thread: Optional[Thread] = None):
		if thread is None:
			thread = current_thread()
		with self.__lock:
			return self.__get_dict(thread).get(key, default)

	def put(self, key: Any, value: Any, *, thread: Optional[Thread] = None):
		if thread is None:
			thread = current_thread()
		with self.__lock:
			self.__get_dict(thread)[key] = value

	def pop(self, key, *, thread: Optional[Thread] = None) -> Optional[Any]:
		if thread is None:
			thread = current_thread()
		with self.__lock:
			dt = self.__get_dict(thread)
			if key in dt:
				return dt.pop(key)
			else:
				return None



import json
import os
from typing import Optional, Dict, TYPE_CHECKING, Union

from mcdreforged.command.command_source import CommandSource, PlayerCommandSource, ConsoleCommandSource
from mcdreforged.constants import core_constant, plugin_constant
from mcdreforged.utils.serializer import Serializable, deserialize, serialize

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer


PREFERENCE_FILE = os.path.join(plugin_constant.PLUGIN_CONFIG_DIRECTORY, core_constant.PACKAGE_NAME, 'preferences.json')
CONSOLE_ALIAS = '#@MCDR_Console@#'


class PreferenceItem(Serializable):
	language: Optional[str]


PreferenceStorage = Dict[str, PreferenceItem]


class PreferenceManager:
	def __init__(self, mcdr_server: 'MCDReforgedServer'):
		self.mcdr_server: 'MCDReforgedServer' = mcdr_server
		self.logger = mcdr_server.logger
		self.preferences: PreferenceStorage = {}

	def load_preferences(self):
		try:
			with open(PREFERENCE_FILE, 'r', encoding='utf8') as file:
				self.preferences = deserialize(json.load(file), PreferenceStorage, error_at_missing=True, error_at_redundancy=True)
		except Exception as e:
			if not isinstance(e, FileNotFoundError):
				self.logger.exception('Failed to load preference file')
			self.preferences = {}
			self.save_preferences()

	def save_preferences(self):
		try:
			dir_path = os.path.dirname(PREFERENCE_FILE)
			if not os.path.isdir(dir_path):
				os.makedirs(dir_path)
			with open(PREFERENCE_FILE, 'w', encoding='utf8') as file:
				json.dump(serialize(self.preferences), file, indent=4, ensure_ascii=False)
		except:
			self.logger.exception('Failed to save preference file')

	def get_default_preference(self) -> PreferenceItem:
		return PreferenceItem(
			language=self.mcdr_server.get_language()
		)

	def get_preference(self, obj: Union[str, CommandSource], *, auto_add: bool = False, strict_type_check: bool = False) -> PreferenceItem:
		if isinstance(obj, str):
			player_name = obj
		elif isinstance(obj, PlayerCommandSource):
			player_name = obj.player
		elif isinstance(obj, ConsoleCommandSource):
			player_name = CONSOLE_ALIAS
		else:
			player_name = None
			if strict_type_check:
				raise TypeError('Unsupported object type during preference querying: {}'.format(type(obj)))
		pref = self.preferences.get(player_name)
		if pref is None:
			pref = self.get_default_preference()
			if auto_add and player_name is not None:
				self.preferences[player_name] = pref
				self.save_preferences()
		return pref

	def get_preferred_language(self, obj: Union[str, CommandSource]) -> str:
		pref = self.get_preference(obj)
		if pref is not None:
			return pref.language
		return self.mcdr_server.translation_manager.language





import os
from typing import List, TYPE_CHECKING

from mcdreforged.minecraft.rtext import RTextList, RText, RAction, RTextBase

if TYPE_CHECKING:
	from mcdreforged.plugin.type.plugin import AbstractPlugin
	from mcdreforged.plugin.plugin_manager import PluginManager


class SingleOperationResult:
	def __init__(self):
		self.success_list = []  # type: List['AbstractPlugin']
		self.failed_list = []  # type: List['AbstractPlugin' or str]

	def succeed(self, plugin: 'AbstractPlugin'):
		self.success_list.append(plugin)

	def fail(self, plugin: 'AbstractPlugin' or str):
		self.failed_list.append(plugin)

	def record(self, plugin: 'AbstractPlugin' or str, result: bool):
		if result:
			self.succeed(plugin)
		else:
			self.fail(plugin)

	def clear(self):
		self.success_list.clear()
		self.failed_list.clear()

	def has_success(self):
		return len(self.success_list) > 0

	def has_failed(self):
		return len(self.failed_list) > 0


class PluginOperationResult:
	def __init__(self, plugin_manager: 'PluginManager'):
		self.__plugin_manager = plugin_manager
		self.__mcdr_server = plugin_manager.mcdr_server
		self.load_result = SingleOperationResult()
		self.unload_result = SingleOperationResult()
		self.reload_result = SingleOperationResult()
		self.dependency_check_result = SingleOperationResult()

	def record(self, load_result, unload_result, reload_result, dependencies_resolve_result):
		self.load_result = load_result
		self.unload_result = unload_result
		self.reload_result = reload_result
		self.dependency_check_result = dependencies_resolve_result

	def clear(self):
		self.load_result.clear()
		self.unload_result.clear()
		self.reload_result.clear()
		self.dependency_check_result.clear()

	def to_rtext(self, *, show_path: bool) -> RTextBase:
		def add_element(msg: RTextList, element):
			msg.append(element)
			msg.append('; ')

		def add_if_not_empty(msg: RTextList, lst: List['AbstractPlugin' or str], key: str):
			if len(lst) > 0:
				text_list = []
				for ele in lst:
					if isinstance(ele, str):
						text_list.append(ele if show_path else os.path.basename(ele))
					else:
						text_list.append(str(ele))
				add_element(msg, RText(self.__mcdr_server.tr(key, len(lst))).h('\n'.join(text_list)))

		message = RTextList()
		add_if_not_empty(message, list(filter(lambda plg: plg in self.dependency_check_result.success_list, self.load_result.success_list)), 'plugin_operation_result.info_loaded_succeeded')
		add_if_not_empty(message, self.unload_result.success_list, 'plugin_operation_result.info_unloaded_succeeded')
		add_if_not_empty(message, self.reload_result.success_list, 'plugin_operation_result.info_reloaded_succeeded')
		add_if_not_empty(message, self.load_result.failed_list, 'plugin_operation_result.info_loaded_failed')
		add_if_not_empty(message, self.unload_result.failed_list, 'plugin_operation_result.info_unloaded_failed')
		add_if_not_empty(message, self.reload_result.failed_list, 'plugin_operation_result.info_reloaded_failed')
		add_if_not_empty(message, self.dependency_check_result.failed_list, 'plugin_operation_result.info_dependency_check_failed')
		if message.is_empty():
			add_element(message, self.__mcdr_server.tr('plugin_operation_result.info_none'))
		message.append(
			RText(self.__mcdr_server.tr('plugin_operation_result.info_plugin_amount', self.__plugin_manager.get_plugin_amount())).
				h('\n'.join(map(str, self.__plugin_manager.get_all_plugins()))).
				c(RAction.suggest_command, '!!MCDR plugin list')
		)
		return message



"""
For reacting general info
Including on_info and !!MCDR, !!help command
"""

from mcdreforged.info_reactor.abstract_info_reactor import AbstractInfoReactor
from mcdreforged.plugin.plugin_event import MCDRPluginEvents


class GeneralReactor(AbstractInfoReactor):
	def react(self, info):
		command_source = info.get_command_source()
		if command_source is not None:
			self.mcdr_server.command_manager.execute_command(info.content, command_source)

		self.mcdr_server.plugin_manager.dispatch_event(MCDRPluginEvents.GENERAL_INFO, (info,))

		if info.is_user:
			self.mcdr_server.plugin_manager.dispatch_event(MCDRPluginEvents.USER_INFO, (info,))



"""
Single plugin class
"""
from enum import Enum, auto
from typing import Tuple, Any, TYPE_CHECKING

from mcdreforged.command.builder.nodes.basic import Literal
from mcdreforged.plugin.meta.metadata import Metadata
from mcdreforged.plugin.plugin_event import MCDREvent, EventListener, PluginEvent
from mcdreforged.plugin.plugin_registry import PluginRegistry, HelpMessage
from mcdreforged.utils.exception import IllegalCallError, IllegalStateError
from mcdreforged.utils.logger import DebugOption
from mcdreforged.utils.types import TranslationKeyDictNested

if TYPE_CHECKING:
	from mcdreforged.plugin.plugin_manager import PluginManager


class PluginState(Enum):
	UNINITIALIZED = auto()  # just created the instance
	LOADING = auto()        # loading the .py entrance file
	LOADED = auto()         # loaded the .py entrance file
	READY = auto()          # called "on load" event, ready to do anything
	UNLOADING = auto()      # just removed from the plugin list, ready to call "on unload" event
	UNLOADED = auto()       # unloaded, should never access it


class AbstractPlugin:
	def __init__(self, plugin_manager: 'PluginManager', file_path: str):
		self.plugin_manager = plugin_manager
		self.mcdr_server = plugin_manager.mcdr_server
		self.plugin_path = file_path
		self.state = PluginState.UNINITIALIZED
		self.plugin_registry = PluginRegistry(self, plugin_manager.registry_storage)

		from mcdreforged.plugin.server_interface import PluginServerInterface
		self.server_interface = PluginServerInterface(self.mcdr_server, self)

	def is_permanent(self) -> bool:
		return False

	def is_regular(self) -> bool:
		return False

	def get_metadata(self) -> Metadata:
		raise NotImplementedError()

	def get_id(self) -> str:
		return self.get_metadata().id

	def get_meta_name(self) -> str:
		return self.get_metadata().name

	def get_fallback_metadata_id(self) -> str:
		raise NotImplementedError()

	def get_name(self) -> str:
		try:
			return self.get_identifier()
		except IllegalCallError:
			return repr(self)

	def get_identifier(self) -> str:
		meta_data = self.get_metadata()
		return '{}@{}'.format(meta_data.id, meta_data.version)

	def __str__(self):
		return self.get_name()

	def __repr__(self):
		raise NotImplementedError()

	# ----------------
	#   Plugin State
	# ----------------

	def set_state(self, state):
		self.state = state

	def in_states(self, states):
		return self.state in states

	def assert_state(self, states, extra_message=None):
		if not self.in_states(states):
			msg = '{} state assertion failed, excepts {} but founded {}.'.format(repr(self), states, self.state)
			if extra_message is not None:
				msg += ' ' + extra_message
			raise IllegalStateError(msg)

	# --------------
	#   Life Cycle
	# --------------

	def load(self):
		raise NotImplementedError()

	def ready(self):
		raise NotImplementedError()

	def reload(self):
		raise NotImplementedError()

	def unload(self):
		raise NotImplementedError()

	def remove(self):
		raise NotImplementedError()

	# -------------------
	#   Plugin Registry
	# -------------------

	def __assert_allow_to_register(self, target):
		self.assert_state([PluginState.LOADED, PluginState.READY], 'Only plugin in loaded or ready state is allowed to register {}'.format(target))

	def register_event_listener(self, event: PluginEvent, listener: EventListener):
		self.__assert_allow_to_register('listener')
		self.plugin_registry.register_event_listener(event.id, listener)
		self.mcdr_server.logger.debug('{} is registered for {}'.format(listener, event), option=DebugOption.PLUGIN)

	def register_command(self, node: Literal):
		self.__assert_allow_to_register('command')
		self.plugin_registry.register_command(node)
		self.mcdr_server.logger.debug('{} registered command with root node {}'.format(self, node), option=DebugOption.PLUGIN)

	def register_help_message(self, help_message: HelpMessage):
		self.__assert_allow_to_register('help message')
		self.plugin_registry.register_help_message(help_message)
		self.mcdr_server.logger.debug('{} registered help message "{}"'.format(self, help_message), option=DebugOption.PLUGIN)

	def receive_event(self, event: MCDREvent, args: Tuple[Any, ...]):
		"""
		Directly dispatch an event towards this plugin
		Not suggested to invoke directly in general case since it doesn't have priority control
		"""
		self.assert_state({PluginState.READY, PluginState.UNLOADING}, 'Only plugin in READY or UNLOADING state is allowed to receive events')
		self.mcdr_server.logger.debug('{} directly received {}'.format(self, event), option=DebugOption.PLUGIN)
		for listener in self.plugin_registry.event_listeners.get(event.id, []):
			self.plugin_manager.trigger_listener(listener, args)

	def register_translation(self, language: str, mapping: TranslationKeyDictNested):
		self.__assert_allow_to_register('translation')
		self.plugin_registry.register_translation(language, mapping)
		self.mcdr_server.logger.debug('{} registered translation for {} with at least {} entries'.format(self, language, len(mapping)), option=DebugOption.PLUGIN)



import os
from logging import Logger
from threading import RLock

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap

from mcdreforged.utils import resources_util
from mcdreforged.utils.lazy_item import LazyItem


class YamlDataStorage:
	def __init__(self, logger: Logger, file_path: str, default_file_path: str):
		self.logger = logger
		self.__file_path = file_path
		self.__default_file_path = default_file_path
		self._data = CommentedMap()
		self.__default_data = LazyItem(lambda: resources_util.get_yaml(self.__default_file_path))
		self.__has_changes = False
		self._data_operation_lock = RLock()

	def to_dict(self) -> dict:
		def process(data: dict) -> dict:
			ret = {}
			for key, value in data.items():
				if isinstance(value, dict):
					value = process(value)
				ret[key] = value
			return ret
		return process(self._data)

	def file_presents(self) -> bool:
		return os.path.isfile(self.__file_path)

	def _load_data(self, allowed_missing_file) -> bool:
		"""
		:param bool allowed_missing_file: If set to True, missing data file will result in a FileNotFoundError(),
		otherwise it will treat it as an empty config file
		:return: if there is any missing data entry
		:raise: FileNotFoundError
		"""
		if self.file_presents():
			with open(self.__file_path, encoding='utf8') as file:
				users_data = YAML().load(file)
		else:
			if not allowed_missing_file:
				raise FileNotFoundError()
			users_data = {}
		self.__has_changes = False
		fixed_result = self.__fix(dict(self.__default_data.get()), users_data)
		with self._data_operation_lock:
			self._data = fixed_result
		if self.__has_changes:
			self.save()
		return self.__has_changes

	def __fix(self, current_data: dict, users_data: CommentedMap, key_path='') -> dict:
		if not isinstance(users_data, dict):
			self.__has_changes = True
			return current_data
		else:
			result = users_data.copy()
			divider = ' -> ' if len(key_path) > 0 else ''
			for key in current_data.keys():
				current_key_path = key_path + divider + key
				if key in users_data:
					# if key presents in user's data
					if isinstance(current_data[key], dict):
						# dive deeper
						result[key] = self.__fix(current_data[key], users_data[key], current_key_path)
					else:
						# use the value in user's data
						result[key] = users_data[key]
				else:
					# missing config key
					result[key] = current_data[key]
					self.__has_changes = True
					self.logger.warning('Option "{}" missing, use default value "{}"'.format(current_key_path, current_data[key]))
			return result

	def _pre_save(self, data: CommentedMap):
		pass

	def __save(self, data: CommentedMap):
		with self._data_operation_lock:
			self._pre_save(data)
			with open(self.__file_path, 'w', encoding='utf8') as file:
				YAML().dump(data, file)

	def save(self):
		self.__save(self._data)

	def get_default_yaml(self):
		return self.__default_data.get()

	def save_default(self):
		self.__save(self.get_default_yaml())



from typing import Union, Dict

from mcdreforged.minecraft.rtext import RTextBase

MessageText = Union[str, RTextBase]
TranslationLanguageDict = Dict[str, str]  # language -> text
TranslationKeyDict = Dict[str, str]  # key -> text
TranslationKeyDictNested = Dict[str, Union[str, 'TranslationKeyDictNested']]  # key -> text/(key -> text/(...))
TranslationKeyDictRich = Dict[str, MessageText]  # language -> text
TranslationStorage = Dict[str, TranslationLanguageDict]  # key -> (language -> text)



from abc import ABC
from typing import Optional, Union

from mcdreforged.minecraft.rtext import RTextBase, RText, RColor
from mcdreforged.utils.types import MessageText


class CommandErrorBase(Exception, ABC):
	pass


class IllegalNodeOperation(CommandErrorBase):
	"""
	This node is not allowed to do this
	"""
	pass


class CommandError(CommandErrorBase, ABC):
	"""
	The basic error, for errors raising when a command source is executing a command
	"""
	def __init__(self, message: MessageText, parsed_command: str, failed_command: str):
		#  !!something wroooong command
		#  [--parsed--|-error-]
		#  [------failed------]

		self.__message: MessageText = message
		self._parsed_command: str = parsed_command
		self._failed_command: str = failed_command
		self.__handled: bool = False

	def __str__(self):
		return '{}: {}<--'.format(self.__message, self._failed_command)

	@property
	def _error_command(self) -> str:
		return self._failed_command[len(self._parsed_command):]

	def to_rtext(self) -> RTextBase:
		return RTextBase.format(
			'{}: {}{}{}',
			RTextBase.from_any(self.__message).copy().set_color(RColor.red),
			RText(self._parsed_command, RColor.dark_red),
			RText(self._error_command, RColor.red),
			RText('<--', RColor.dark_red)
		)

	def get_error_data(self) -> tuple:
		"""
		Data that might be helpful to the error display
		Can be used in formatting processing
		"""
		return ()

	def set_message(self, message: str):
		self.__message = message

	def get_parsed_command(self) -> str:
		return self._parsed_command

	def get_failed_command(self) -> str:
		return self._failed_command

	def set_handled(self) -> None:
		"""
		It won't make any difference to the command node tree execution
		But it might be useful for outer error handlers
		"""
		self.__handled = True

	def is_handled(self) -> bool:
		return self.__handled


class UnknownCommand(CommandError):
	"""
	When the command finishes parsing, but current node doesn't have a callback function
	"""
	def __init__(self, parsed_command, failed_command):
		super().__init__('Unknown Command', parsed_command, failed_command)


class UnknownArgument(CommandError):
	"""
	When there's remaining command string, but there's no matched Literal nodes and no general argument nodes
	"""
	def __init__(self, parsed_command: str, failed_command: str):
		super().__init__('Unknown Argument', parsed_command, failed_command)


class UnknownRootArgument(UnknownArgument):
	"""
	The same as UnknownArgument, but it fails to match at root node
	"""
	pass


class RequirementNotMet(CommandError):
	"""
	The specified requirement for the command source to enter this node is not met
	"""
	__NO_REASON = RText('Requirement not met')

	def __init__(self, parsed_command: str, failed_command: str, reason: Optional[MessageText]):
		self.__reason: MessageText = reason if reason is not None else self.__NO_REASON
		super().__init__(self.__reason, parsed_command, failed_command)

	def has_custom_reason(self) -> bool:
		return self.__reason is not self.__NO_REASON

	def get_reason(self) -> MessageText:
		return self.__reason

	def get_error_data(self) -> tuple:
		return (self.get_reason(),)

# -----------------
#   Syntax things
# -----------------


class CommandSyntaxError(CommandError, ABC):
	"""
	General illegal argument error
	Used in integer parsing failure etc.
	"""
	def __init__(self, message: str, char_read: Union[int, str]):
		super().__init__(message, '', '?' if isinstance(char_read, int) else char_read)
		self.message = message
		self.char_read = char_read if isinstance(char_read, int) else len(char_read)

	def set_parsed_command(self, parsed_command):
		self._parsed_command = parsed_command

	def set_failed_command(self, failed_command):
		self._failed_command = failed_command


class IllegalArgument(CommandSyntaxError, ABC):
	"""
	General illegal argument error
	Used in integer parsing failure etc.
	"""
	pass


class LiteralNotMatch(CommandSyntaxError):
	"""
	Used by Literal node parsing failure for fail-soft
	"""
	pass


class AbstractOutOfRange(IllegalArgument, ABC):
	def __init__(self, message: str, char_read: Union[int, str], value, range_l, range_r):
		"""
		:param value: The actual value
		:param range_l: The left boundary
		:param range_r: The right boundary
		"""
		super().__init__(message, char_read)
		self.__value = value
		self.__range_l = range_l
		self.__range_r = range_r

	@classmethod
	def _get_boundary_text(cls, value) -> str:
		return str(value) if value is not None else '/'

	def get_error_data(self) -> tuple:
		return self.__value, self._get_boundary_text(self.__range_l), self._get_boundary_text(self.__range_r)


# Number things


class NumberOutOfRange(AbstractOutOfRange):
	"""
	The parsed number value is out of the restriction range
	"""
	def __init__(self, char_read: Union[int, str], value, range_l, range_r):
		super().__init__('Value out of range [{}, {}]'.format(self._get_boundary_text(range_l), self._get_boundary_text(range_r)), char_read, value, range_l, range_r)


class InvalidNumber(IllegalArgument):
	def __init__(self, char_read: Union[int, str]):
		super().__init__('Invalid number', char_read)


class InvalidInteger(IllegalArgument):
	def __init__(self, char_read: Union[int, str]):
		super().__init__('Invalid integer', char_read)


class InvalidFloat(IllegalArgument):
	def __init__(self, char_read: Union[int, str]):
		super().__init__('Invalid float', char_read)


# Text things


class TextLengthOutOfRange(AbstractOutOfRange):
	"""
	The length of the given text is out of the restriction range
	"""
	def __init__(self, char_read: Union[int, str], value, range_l, range_r):
		super().__init__('Text length {} out of range [{}, {}]'.format(value, self._get_boundary_text(range_l), self._get_boundary_text(range_r)), char_read, value, range_l, range_r)


class IllegalEscapesUsage(IllegalArgument):
	"""
	The text is empty, and it's not allowed to be
	"""
	def __init__(self, char_read: Union[int, str]):
		super().__init__('Illegal usage of escapes', char_read)


class UnclosedQuotedString(IllegalArgument):
	"""
	The text is empty, and it's not allowed to be
	"""
	def __init__(self, char_read: Union[int, str]):
		super().__init__('Unclosed quoted string', char_read)


class EmptyText(IllegalArgument):
	"""
	The text is empty, and it's not allowed to be
	"""
	def __init__(self, char_read: Union[int, str]):
		super().__init__('Empty text is not allowed', char_read)


# Other Arguments


class InvalidBoolean(IllegalArgument):
	def __init__(self, char_read: Union[int, str]):
		super().__init__('Invalid boolean', char_read)


class InvalidEnumeration(IllegalArgument):
	def __init__(self, char_read: Union[int, str]):
		super().__init__('Invalid enumeration', char_read)



import copy
from typing import Optional


class ServerInformation:
	version: Optional[str] = None
	ip: Optional[str] = None
	port: Optional[int] = None

	def clear(self):
		self.version, self.ip, self.port = None, None, None

	def copy(self) -> 'ServerInformation':
		return copy.copy(self)




import os
from typing import Callable


def list_all(directory: str, filter: Callable[[str], bool] = lambda file_path: True):
	ret = []
	for file in os.listdir(directory):
		file_path = os.path.join(directory, file)
		if filter(file_path):
			ret.append(file_path)
	return ret


def list_file(directory: str, filter: Callable[[str], bool] = lambda file_path: True):
	return list_all(directory, lambda file_path: os.path.isfile(file_path) and filter(file_path))


def list_file_with_suffix(directory: str, suffix: str):
	return list_file(directory, lambda file_path: file_path.endswith(suffix))


def touch_directory(directory_path: str):
	if not os.path.isdir(directory_path):
		os.makedirs(directory_path)


def get_file_suffix(file_path: str):
	file_name = os.path.basename(file_path)
	index = file_name.rfind('.')
	if index == -1:
		return ''
	return file_name[index:]



"""
Handling MCDR commands
"""
import collections
from enum import Enum, auto
from typing import TYPE_CHECKING, Dict, List

import mcdreforged.command.builder.command_builder_util as utils
from mcdreforged.command.builder.exception import CommandError, RequirementNotMet
from mcdreforged.command.builder.nodes.basic import CommandSuggestion, CommandSuggestions
from mcdreforged.command.command_source import InfoCommandSource, CommandSource
from mcdreforged.plugin.plugin_registry import PluginCommandNode
from mcdreforged.utils import string_util
from mcdreforged.utils.logger import DebugOption

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer


class TraversePurpose(Enum):
	EXECUTE = auto()
	SUGGEST = auto()


# deal with !!MCDR and !!help command
class CommandManager:
	def __init__(self, mcdr_server: 'MCDReforgedServer'):
		self.mcdr_server = mcdr_server
		self.logger = self.mcdr_server.logger
		self.tr = self.mcdr_server.tr
		self.root_nodes = collections.defaultdict(list)  # type: Dict[str, List[PluginCommandNode]]

		self.__preserve_command_error_display_flag = False

	def clear_command(self):
		self.root_nodes.clear()

	def register_command(self, plugin_node: PluginCommandNode):
		for literal in plugin_node.node.literals:
			self.root_nodes[literal].append(plugin_node)

	def _traverse(self, command: str, source: CommandSource, purpose: TraversePurpose) -> None or List[CommandSuggestion]:
		def __translate_command_error_header(translation_key_: str, error_: CommandError) -> str:
			if isinstance(error_, RequirementNotMet):
				if error_.has_custom_reason():
					return error_.get_reason()
				args = ()
			else:
				args = error_.get_error_data()
			return self.mcdr_server.tr(translation_key_, *args, allow_failure=False, language=source.get_preference().language)

		first_literal_element = utils.get_element(command)
		plugin_root_nodes = self.root_nodes.get(first_literal_element, [])
		suggestions = CommandSuggestions()

		if purpose == TraversePurpose.EXECUTE and isinstance(source, InfoCommandSource):
			if len(plugin_root_nodes) > 0 and source.is_console:
				# If this is a command, don't send it towards the server if it's from console input
				source.get_info().cancel_send_to_server()

		if purpose == TraversePurpose.SUGGEST and len(plugin_root_nodes) == 0:
			return CommandSuggestions([CommandSuggestion('', literal) for literal in self.root_nodes.keys()])

		for plugin_root_node in plugin_root_nodes:
			plugin = plugin_root_node.plugin
			node = plugin_root_node.node
			try:
				with self.mcdr_server.plugin_manager.with_plugin_context(plugin):
					if purpose == TraversePurpose.EXECUTE:
						node.execute(source, command)
					elif purpose == TraversePurpose.SUGGEST:
						suggestions.extend(node.generate_suggestions(source, command))

			except CommandError as error:
				if not error.is_handled():
					translation_key = 'command_exception.{}'.format(string_util.hump_to_underline(type(error).__name__))
					try:
						error.set_message(__translate_command_error_header(translation_key, error))
					except KeyError:
						self.logger.debug('Fail to translated command error with key {}'.format(translation_key), option=DebugOption.COMMAND)
					source.reply(error.to_rtext())
			except:
				self.logger.exception('Error when executing command "{}" with command source "{}" on {} registered by {}'.format(command, source, node, plugin))

		if purpose == TraversePurpose.SUGGEST:
			return suggestions

	def execute_command(self, command: str, source: CommandSource):
		self._traverse(command, source, TraversePurpose.EXECUTE)

	def suggest_command(self, command: str, source: CommandSource) -> CommandSuggestions:
		return self._traverse(command, source, TraversePurpose.SUGGEST)



from mcdreforged.handler.impl.basic_handler import BasicHandler
from mcdreforged.handler.impl.beta18_handler import Beta18Handler
from mcdreforged.handler.impl.bukkit14_handler import Bukkit14Handler
from mcdreforged.handler.impl.bukkit_handler import BukkitHandler
from mcdreforged.handler.impl.bungeecord_handler import BungeecordHandler
from mcdreforged.handler.impl.cat_server_handler import CatServerHandler
from mcdreforged.handler.impl.forge_handler import ForgeHandler
from mcdreforged.handler.impl.vanilla_handler import VanillaHandler
from mcdreforged.handler.impl.velocity_handler import VelocityHandler
from mcdreforged.handler.impl.waterfall_handler import WaterfallHandler

__all__ = [
	'BasicHandler',

	'VanillaHandler',
	'BukkitHandler', 'Bukkit14Handler', 'CatServerHandler',
	'ForgeHandler',
	'Beta18Handler',

	'BungeecordHandler',
	'WaterfallHandler',
	'VelocityHandler'
]



from mcdreforged.handler.impl.bukkit_handler import BukkitHandler
from mcdreforged.handler.impl.vanilla_handler import VanillaHandler
from mcdreforged.utils import string_util


class CatServerHandler(BukkitHandler):
	# https://github.com/Luohuayu/CatServer
	# CatServer uses vanilla logging format but spigot like player joined message
	# And has color code around the player left message

	@staticmethod
	def __cleaned_info(info):
		info.content = string_util.clean_minecraft_color_code(info.content)
		return info

	@classmethod
	def get_content_parsing_formatter(cls):
		return VanillaHandler.get_content_parsing_formatter()

	def parse_player_left(self, info):
		# §eSteve left the game§r
		return super().parse_player_left(self.__cleaned_info(info))



from typing import Optional

from parse import parse

from mcdreforged.handler.impl.bukkit_handler import BukkitHandler
from mcdreforged.handler.impl.bungeecord_handler import BungeecordHandler
from mcdreforged.info_reactor.info import Info


class WaterfallHandler(BungeecordHandler):
	# The logging format of waterfall server is paper like
	# [02:18:30 INFO]: Enabled plugin cmd_list version git:cmd_list:1.15-SNAPSHOT:f1c32f8:1489 by SpigotMC
	# [02:18:29 INFO] [ViaVersion]: Loading 1.12.2 -> 1.13 mappings..."

	@classmethod
	def get_content_parsing_formatter(cls):
		return (
			BukkitHandler.get_content_parsing_formatter(),
			'[{hour:d}:{min:d}:{sec:d} {logging}] {dummy}: {content}'  # somethings there is an extra element after the heading [] and :
		)

	def parse_player_joined(self, info: Info) -> Optional[str]:
		# [02:18:52 INFO]: [/127.0.0.1:14426] <-> InitialHandler has connected
		# sadly no player id display here
		return None

	def parse_player_left(self, info):
		# [/127.0.0.1:14426|Fallen_Breath] -> UpstreamBridge has disconnected
		if not info.is_user:
			parsed = parse('[/{ip}|{name}] -> UpstreamBridge has disconnected', info.content)
			if parsed is not None:
				return parsed['name']
		return None



"""
Plugin management
"""
import collections
import os
import threading
from contextlib import contextmanager
from typing import Callable, Dict, Optional, Any, Tuple, List, TYPE_CHECKING, Deque

from mcdreforged.constants import core_constant, plugin_constant
from mcdreforged.plugin import plugin_factory
from mcdreforged.plugin.builtin.mcdreforged_plugin.mcdreforged_plugin import MCDReforgedPlugin
from mcdreforged.plugin.builtin.python_plugin import PythonPlugin
from mcdreforged.plugin.meta.dependency_walker import DependencyWalker
from mcdreforged.plugin.operation_result import PluginOperationResult, SingleOperationResult
from mcdreforged.plugin.plugin_event import MCDRPluginEvents, MCDREvent, EventListener
from mcdreforged.plugin.plugin_registry import PluginRegistryStorage
from mcdreforged.plugin.plugin_thread import PluginThreadPool
from mcdreforged.plugin.type.permanent_plugin import PermanentPlugin
from mcdreforged.plugin.type.plugin import AbstractPlugin, PluginState
from mcdreforged.plugin.type.regular_plugin import RegularPlugin
from mcdreforged.utils import file_util, string_util, misc_util
from mcdreforged.utils.logger import DebugOption
from mcdreforged.utils.thread_local_storage import ThreadLocalStorage

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer


class PluginManager:
	TLS_PLUGIN_KEY = 'current_plugin'

	def __init__(self, mcdr_server: 'MCDReforgedServer'):
		self.plugin_directories = []  # type: List[str]
		self.mcdr_server = mcdr_server
		self.logger = mcdr_server.logger

		# id -> Plugin plugin storage
		self.plugins = {}   # type: Dict[str, AbstractPlugin]
		# file_path -> id mapping
		self.plugin_file_path = {}  # type: Dict[str, str]
		# storage for event listeners, help messages and commands
		self.registry_storage = PluginRegistryStorage(self)

		self.last_operation_result = PluginOperationResult(self)

		# not used currently
		self.thread_pool = PluginThreadPool(self.mcdr_server, max_thread=core_constant.PLUGIN_THREAD_POOL_SIZE)

		# thread local storage, to store current plugin
		self.__tls = ThreadLocalStorage()

		# plugin manipulation lock
		self.__mani_lock = threading.RLock()

	@classmethod
	def touch_directory(cls):
		file_util.touch_directory(plugin_constant.PLUGIN_CONFIG_DIRECTORY)

	# --------------------------
	#   Getters / Setters etc.
	# --------------------------

	def get_current_running_plugin(self, *, thread=None) -> Optional[AbstractPlugin]:
		"""
		Get current executing plugin in this thread
		:param thread: If specified, it should be a Thread instance. Then it will return the executing plugin in the given thread
		"""
		stack = self.__tls.get(self.TLS_PLUGIN_KEY, None, thread=thread)  # type: Deque[AbstractPlugin]
		return stack[-1] if stack is not None else None

	def get_plugin_amount(self) -> int:
		return len(self.plugins)

	def get_all_plugins(self) -> List[AbstractPlugin]:
		return list(self.plugins.values())

	def get_regular_plugins(self) -> List[RegularPlugin]:
		return [plugin for plugin in self.plugins.values() if isinstance(plugin, RegularPlugin)]

	def get_plugin_from_id(self, plugin_id: str) -> Optional[AbstractPlugin]:
		return self.plugins.get(plugin_id)

	def get_regular_plugin_from_id(self, plugin_id: str) -> Optional[RegularPlugin]:
		plugin = self.get_plugin_from_id(plugin_id)
		if not isinstance(plugin, RegularPlugin):
			plugin = None
		return plugin

	def set_plugin_directories(self, plugin_directories: Optional[List[str]]):
		if plugin_directories is None:
			plugin_directories = []
		self.plugin_directories = misc_util.unique_list(plugin_directories)
		for plugin_directory in self.plugin_directories:
			file_util.touch_directory(plugin_directory)

	@contextmanager
	def with_plugin_context(self, plugin: AbstractPlugin):
		stack = self.__tls.get(self.TLS_PLUGIN_KEY, default=collections.deque())  # type: Deque[AbstractPlugin]
		stack.append(plugin)
		self.__tls.put(self.TLS_PLUGIN_KEY, stack)
		try:
			yield
		finally:
			stack.pop()
			if len(stack) == 0:
				self.__tls.pop(self.TLS_PLUGIN_KEY)

	def contains_plugin_file(self, file_path: str) -> bool:
		return os.path.abspath(file_path) in self.plugin_file_path

	def contains_plugin_id(self, plugin_id: str) -> bool:
		"""
		Includes permanent plugins
		"""
		return plugin_id in self.plugins

	# ---------------------------------------
	#   Permanent build-in plugin operation
	# ---------------------------------------

	def __add_permanent_plugin(self, plugin: PermanentPlugin):
		self.__add_plugin(plugin)
		plugin.set_state(PluginState.LOADED)
		plugin.load()
		plugin.set_state(PluginState.READY)

	def register_permanent_plugins(self):
		self.__add_permanent_plugin(MCDReforgedPlugin(self))
		self.__add_permanent_plugin(PythonPlugin(self))
		self.__sort_plugins_by_id()
		self.__update_registry()

	# ------------------------------------------------
	#   Actual operations that add / remove a plugin
	# ------------------------------------------------

	def __add_plugin(self, plugin: AbstractPlugin):
		plugin_id = plugin.get_id()
		if plugin_id in self.plugins:
			self.logger.critical('Something is not correct, a plugin with existed plugin id "{}" is added'.format(plugin_id))
		self.plugins[plugin_id] = plugin
		if isinstance(plugin, RegularPlugin):
			self.plugin_file_path[os.path.abspath(plugin.plugin_path)] = plugin_id

	def __remove_plugin(self, plugin: AbstractPlugin):
		if not plugin.is_permanent():
			plugin_id = plugin.get_id()
			self.plugins.pop(plugin_id, None)
			if isinstance(plugin, RegularPlugin):
				self.plugin_file_path.pop(os.path.abspath(plugin.plugin_path), None)

	# ----------------------------
	#   Single Plugin Operations
	# ----------------------------

	def __load_plugin(self, file_path):
		"""
		Try to load a plugin from the given file
		If succeeds, add the plugin to the plugin list, the plugin state will be set to LOADED
		If fails, nothing will happen
		:param str file_path: The path to the plugin file, a *.py
		:return: the new plugin instance if succeeds, otherwise None
		:rtype: RegularPlugin or None
		"""
		plugin = plugin_factory.create_regular_plugin(self, file_path)
		try:
			plugin.load()
		except:
			self.logger.exception(self.mcdr_server.tr('plugin_manager.load_plugin.fail', plugin.get_name()))
			return None
		else:
			existed_plugin = self.plugins.get(plugin.get_id())
			if existed_plugin is None:
				self.__add_plugin(plugin)
				self.logger.info(self.mcdr_server.tr('plugin_manager.load_plugin.success', plugin.get_name()))
				return plugin
			else:
				self.logger.error(self.mcdr_server.tr('plugin_manager.load_plugin.duplicate', plugin.get_name(), plugin.plugin_path, existed_plugin.get_name(), existed_plugin.plugin_path))
				try:
					plugin.unload()
				except:
					# should never come here
					self.logger.exception(self.mcdr_server.tr('plugin_manager.load_plugin.unload_duplication_fail', plugin.get_name(), plugin.plugin_path))
				plugin.remove()  # quickly remove this plugin
				return None

	def __unload_plugin(self, plugin):
		"""
		Try to load a plugin from the given file
		Whether it succeeds or not, the plugin instance will be removed from the plugin list
		The plugin state will be set to UNLOADING
		:param AbstractPlugin plugin: The plugin instance to be unloaded
		:return: If there's an exception during plugin unloading
		:rtype: bool
		"""
		try:
			plugin.unload()
		except:
			# should never come here
			plugin.set_state(PluginState.UNLOADING)  # a fallback set state
			self.logger.exception(self.mcdr_server.tr('plugin_manager.unload_plugin.fail', plugin.get_name()))
			ret = False
		else:
			self.logger.info(self.mcdr_server.tr('plugin_manager.unload_plugin.success', plugin.get_name()))
			ret = True
		finally:
			self.__remove_plugin(plugin)
		return ret

	def __reload_plugin(self, plugin):
		"""
		Try to reload an existed plugin
		If fails, unload the plugin and then the plugin state will be set to UNLOADED
		:param AbstractPlugin plugin: The plugin instance to be reloaded
		:return: If the plugin reloads successfully without error
		:rtype: bool
		"""
		plugin.receive_event(MCDRPluginEvents.PLUGIN_UNLOADED, ())
		self.__remove_plugin(plugin)
		try:
			plugin.reload()
		except:
			self.logger.exception(self.mcdr_server.tr('plugin_manager.reload_plugin.fail', plugin.get_name()))
			self.__unload_plugin(plugin)
			return False
		else:
			# in case the plugin id changes into an existed plugin id
			existed_plugin = self.plugins.get(plugin.get_id())
			if existed_plugin is None:
				self.__add_plugin(plugin)
				self.logger.info(self.mcdr_server.tr('plugin_manager.reload_plugin.success', plugin.get_name()))
				return True
			else:
				self.logger.error(self.mcdr_server.tr('plugin_manager.load_plugin.duplicate', plugin.get_name(), plugin.plugin_path, existed_plugin.get_name(), existed_plugin.plugin_path))
				try:
					plugin.unload()
				except:
					# should never come here
					self.logger.exception(self.mcdr_server.tr('plugin_manager.load_plugin.unload_duplication_fail', plugin.get_name(), plugin.plugin_path))
				return False

	# ---------------------------------------
	#   Regular Plugin Collector & Handlers
	# ---------------------------------------

	def __collect_and_process_new_plugins(self, filter: Callable[[str], bool]) -> SingleOperationResult:
		result = SingleOperationResult()
		for plugin_directory in self.plugin_directories:
			if os.path.isdir(plugin_directory):
				for file in os.listdir(plugin_directory):
					file_path = os.path.join(plugin_directory, file)
					if plugin_factory.is_plugin(file_path):
						if not self.contains_plugin_file(file_path) and filter(file_path):
							plugin = self.__load_plugin(file_path)
							if plugin is None:
								result.fail(file_path)
							else:
								result.succeed(plugin)
			else:
				self.logger.warning('Plugin directory "{}" not found'.format(plugin_directory))
		return result

	def __collect_and_remove_plugins(self, filter: Callable[[RegularPlugin], bool], specific: Optional[RegularPlugin] = None) -> SingleOperationResult:
		result = SingleOperationResult()
		plugin_list = self.get_regular_plugins() if specific is None else [specific]
		for plugin in plugin_list:
			if filter(plugin):
				result.record(plugin, self.__unload_plugin(plugin))
		return result

	def __reload_ready_plugins(self, filter: Callable[[RegularPlugin], bool], specific: Optional[RegularPlugin] = None) -> SingleOperationResult:
		result = SingleOperationResult()
		plugin_list = self.get_regular_plugins() if specific is None else [specific]
		for plugin in plugin_list:
			if plugin.in_states({PluginState.READY}) and filter(plugin):
				result.record(plugin, self.__reload_plugin(plugin))
		return result

	def __check_plugin_dependencies(self) -> SingleOperationResult:
		result = SingleOperationResult()
		walker = DependencyWalker(self)
		walk_result = walker.walk()
		for item in walk_result:
			plugin = self.plugins.get(item.plugin_id)  # should be not None
			result.record(plugin, item.success)
			if not item.success:
				self.logger.error(self.mcdr_server.tr('plugin_manager.check_plugin_dependencies.item_failed', plugin, item.reason))
				self.__unload_plugin(plugin)
		self.logger.debug(self.mcdr_server.tr('plugin_manager.check_plugin_dependencies.topo_order'), option=DebugOption.PLUGIN)
		for plugin in result.success_list:
			self.logger.debug('- {}'.format(plugin), option=DebugOption.PLUGIN)
		# the success list order matches the dependency topo order
		return result

	# ---------------------------
	#   Multi-Plugin Operations
	# ---------------------------

	# MCDR 0.x behavior for reload
	# 1. reload existed (and changed) plugins. if reload fail unload it
	# 2. load new plugins, and ignore those plugins which just got unloaded because reloading failed
	# 3. unload removed plugins
	# 4. call on_load for new loaded plugins

	# Current behavior for plugin operations
	# 1. Actual plugin operations
	#   For plugin refresh
	#      1. Unload plugins whose file is removed
	#      2. Load new plugins
	#      3. Reload existed (and matches filter) plugins whose state is ready. if reload fail unload it
	#
	#   For single plugin operation
	#      1. Enable / disable plugin
	#
	# 2. Plugin Processing  (method __post_plugin_process)
	#      1. Check dependencies, unload plugins that has dependencies not satisfied
	#      2. Call on_unload for unloaded plugins
	#      3. Call on_load for new / reloaded plugins by topo order
	#    Dispatch PLUGIN_UNLOADED before PLUGIN_LOADED, so file renamed plugin keeps the same event order (unload -> load) with reloaded plugin

	def __post_plugin_process(self, load_result=None, unload_result=None, reload_result=None):
		"""
		When returns, all events / operations are finished processing
		"""
		if load_result is None:
			load_result = SingleOperationResult()
		if unload_result is None:
			unload_result = SingleOperationResult()
		if reload_result is None:
			reload_result = SingleOperationResult()

		dependency_check_result = self.__check_plugin_dependencies()
		self.last_operation_result.record(load_result, unload_result, reload_result, dependency_check_result)

		# Expected plugin states:
		# 					success_list		fail_list
		# load_result		LOADED				N/A
		# unload_result		UNLOADING			UNLOADING
		# reload_result		READY				UNLOADING
		# dep_chk_result	LOADED / READY		UNLOADING

		self.registry_storage.clear()  # in case plugin invokes dispatch_event during on_load. dont let them trigger listeners

		newly_loaded_plugins = {*load_result.success_list, *reload_result.success_list}
		to_be_removed_plugins = unload_result.success_list + unload_result.failed_list + reload_result.failed_list + dependency_check_result.failed_list

		# collect removed plugin module instance
		removed_plugins_module_instance = {}
		for plugin in to_be_removed_plugins:
			if isinstance(plugin, RegularPlugin):
				removed_plugins_module_instance[plugin.get_id()] = plugin.entry_module_instance

		# get ready
		for plugin in dependency_check_result.success_list:
			if plugin in newly_loaded_plugins:
				plugin.ready()

		# PLUGIN_UNLOADED event
		for plugin in to_be_removed_plugins:
			plugin.assert_state({PluginState.UNLOADING})
			# plugins might just be newly loaded but failed on dependency check, dont dispatch event to them
			if plugin not in newly_loaded_plugins:
				plugin.receive_event(MCDRPluginEvents.PLUGIN_UNLOADED, ())
			plugin.remove()

		# PLUGIN_LOADED event
		for plugin in dependency_check_result.success_list:
			if plugin in newly_loaded_plugins:
				if isinstance(plugin, RegularPlugin):
					old_entry_module_instance = plugin.old_entry_module_instance
					if old_entry_module_instance is None:
						old_entry_module_instance = removed_plugins_module_instance.get(plugin.get_id(), None)
					plugin.receive_event(MCDRPluginEvents.PLUGIN_LOADED, (old_entry_module_instance,))

		# they should all be ready
		for plugin in self.get_regular_plugins():
			plugin.assert_state({PluginState.READY})

		self.__update_registry()
		self.__sort_plugins_by_id()

	def __sort_plugins_by_id(self):
		self.plugins = dict(sorted(self.plugins.items(), key=lambda item: item[0]))

	def __update_registry(self):
		self.registry_storage.clear()
		for plugin in self.get_all_plugins():
			self.registry_storage.collect(plugin.plugin_registry)
		self.registry_storage.arrange()
		self.mcdr_server.on_plugin_registry_changed()

	def __refresh_plugins(self, reload_filter: Callable[[RegularPlugin], bool]):
		unload_result = self.__collect_and_remove_plugins(lambda plugin: not plugin.plugin_exists())
		load_result = self.__collect_and_process_new_plugins(lambda fp: True)
		reload_result = self.__reload_ready_plugins(reload_filter)
		self.__post_plugin_process(load_result, unload_result, reload_result)

	# --------------
	#   Interfaces
	# --------------

	def load_plugin(self, file_path: str):
		with self.__mani_lock:
			self.logger.info(self.mcdr_server.tr('plugin_manager.load_plugin.entered', file_path))
			load_result = self.__collect_and_process_new_plugins(lambda fp: fp == file_path)
			self.__post_plugin_process(load_result=load_result)

	def unload_plugin(self, plugin: RegularPlugin):
		with self.__mani_lock:
			self.logger.info(self.mcdr_server.tr('plugin_manager.unload_plugin.entered', plugin))
			unload_result = self.__collect_and_remove_plugins(lambda plg: True, specific=plugin)
			self.__post_plugin_process(unload_result=unload_result)

	def reload_plugin(self, plugin: RegularPlugin):
		with self.__mani_lock:
			self.logger.info(self.mcdr_server.tr('plugin_manager.reload_plugin.entered', plugin))
			reload_result = self.__reload_ready_plugins(lambda plg: True, specific=plugin)
			self.__post_plugin_process(reload_result=reload_result)

	def enable_plugin(self, file_path: str):
		with self.__mani_lock:
			self.logger.info(self.mcdr_server.tr('plugin_manager.enable_plugin.entered', file_path))
			new_file_path = string_util.remove_suffix(file_path, plugin_constant.DISABLED_PLUGIN_FILE_SUFFIX)
			if plugin_factory.is_disabled_plugin(file_path):
				os.rename(file_path, new_file_path)
				self.load_plugin(new_file_path)

	def disable_plugin(self, plugin: RegularPlugin):
		with self.__mani_lock:
			self.logger.info(self.mcdr_server.tr('plugin_manager.disable_plugin.entered', plugin))
			self.unload_plugin(plugin)
			if plugin.plugin_exists():
				os.rename(plugin.plugin_path, plugin.plugin_path + plugin_constant.DISABLED_PLUGIN_FILE_SUFFIX)

	def refresh_all_plugins(self):
		with self.__mani_lock:
			self.logger.info(self.mcdr_server.tr('plugin_manager.refresh_all_plugins.entered'))
			return self.__refresh_plugins(lambda plg: True)

	def refresh_changed_plugins(self):
		with self.__mani_lock:
			self.logger.info(self.mcdr_server.tr('plugin_manager.refresh_changed_plugins.entered'))
			return self.__refresh_plugins(lambda plg: plg.file_changed())

	# ----------------
	#   Plugin Event
	# ----------------

	def __dispatch_event(self, event: MCDREvent, args: Tuple[Any, ...]):
		"""
		Event dispatch logic implementation
		"""
		self.logger.debug('Dispatching {} with args ({})'.format(event, ', '.join([type(arg).__name__ for arg in args])), option=DebugOption.PLUGIN)
		for listener in self.registry_storage.event_listeners.get(event.id, []):
			self.trigger_listener(listener, args)

	def dispatch_event(self, event: MCDREvent, args: Tuple[Any, ...], *, on_executor_thread: bool = True, block: bool = False, timeout: Optional[float] = None):
		"""
		Event dispatching interface
		"""
		if on_executor_thread:
			self.mcdr_server.task_executor.execute_on_thread(lambda: self.__dispatch_event(event, args), block=block, timeout=timeout)
		else:
			self.__dispatch_event(event, args)

	def trigger_listener(self, listener: EventListener, args: Tuple[Any, ...]):
		"""
		Event listener triggering implementation
		The server_interface parameter will be automatically added as the 1st parameter
		"""
		try:
			with self.with_plugin_context(listener.plugin):
				listener.execute(listener.plugin.server_interface, *args)
		except:
			self.logger.exception('Error invoking listener {}'.format(listener))



from enum import Enum, unique, Flag, auto
from typing import Iterable


@unique
class MCDReforgedFlag(Flag):
	NONE = auto()
	INTERRUPT = auto()			# ctrl-c flag
	SERVER_STARTUP = auto()		# set to True after server startup
	SERVER_RCON_READY = auto() 	# set to True after server started its rcon. used to start the rcon server
	EXIT_AFTER_STOP = auto()	# if MCDR exit after server stop. can be modified by plugins


class EnumStateBase(Enum):
	def in_state(self, states):
		if type(states) is type(self):
			return self is states
		elif not isinstance(states, Iterable):
			states = (states,)
		return self in states


@unique
class MCDReforgedState(EnumStateBase):
	INITIALIZING	= 'mcdr_state.initializing' 	# Just entered construction method
	INITIALIZED		= 'mcdr_state.initialized'		# Construction finished
	RUNNING			= 'mcdr_state.running'			# MCDR started and is running
	PRE_STOPPED		= 'mcdr_state.pre_stopped'		# MCDR is stopping and making some cleaning things
	STOPPED			= 'mcdr_state.stopped'			# MCDR is stopped


@unique
class ServerState(EnumStateBase):
	STOPPED		= 'server_state.stopped'		# Server is stopped
	STOPPING	= 'server_state.stopping'		# Server is being stopped by MCDR
	RUNNING		= 'server_state.running'		# Server is running
	# PRE_STOPPED	= 'server_state.pre_stopped' 	# Server is stopped, and it's going to dispatch_event SERVER_STOP

	def is_server_stopped(self):
		return self.in_state({self.STOPPED})



from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from mcdreforged.mcdr_server import MCDReforgedServer


class AbstractInfoReactor:
	def __init__(self, mcdr_server: 'MCDReforgedServer'):
		self.mcdr_server: 'MCDReforgedServer' = mcdr_server

	def react(self, info):
		raise NotImplementedError()

	def on_server_start(self):
		pass

	def on_server_stop(self):
		pass



import json
from abc import ABC
from enum import Enum
from typing import Type, Iterable, Union

from mcdreforged.command.builder import command_builder_util as utils
from mcdreforged.command.builder.command_builder_util import DIVIDER
from mcdreforged.command.builder.exception import NumberOutOfRange, EmptyText, \
	InvalidNumber, InvalidInteger, InvalidFloat, UnclosedQuotedString, IllegalEscapesUsage, \
	TextLengthOutOfRange, InvalidBoolean, InvalidEnumeration
from mcdreforged.command.builder.nodes.basic import AbstractNode, ParseResult, SUGGESTS_CALLBACK, \
	ArgumentNode, CommandContext
# --------------------
#   Number Arguments
# --------------------
from mcdreforged.utils import misc_util


class NumberNode(ArgumentNode, ABC):
	def __init__(self, name):
		super().__init__(name)
		self.__min_value = None
		self.__max_value = None

	def at_min(self, min_value) -> 'NumberNode':
		self.__min_value = min_value
		return self

	def at_max(self, max_value) -> 'NumberNode':
		self.__max_value = max_value
		return self

	def in_range(self, min_value, max_value) -> 'NumberNode':
		self.at_min(min_value)
		self.at_max(max_value)
		return self

	def _check_in_range_and_return(self, value: Union[int, float], char_read: int):
		if (self.__min_value is not None and value < self.__min_value) or (self.__max_value is not None and value > self.__max_value):
			raise NumberOutOfRange(char_read, value, self.__min_value, self.__max_value)
		return ParseResult(value, char_read)


class Number(NumberNode):
	"""
	An Integer, or a float
	"""
	def parse(self, text: str) -> ParseResult:
		value, read = utils.get_int(text)
		if value is None:
			value, read = utils.get_float(text)
		if value is not None:
			return self._check_in_range_and_return(value, read)
		else:
			raise InvalidNumber(read)


class Integer(NumberNode):
	"""
	An Integer
	"""
	def parse(self, text: str) -> ParseResult:
		value, read = utils.get_int(text)
		if value is not None:
			return self._check_in_range_and_return(value, read)
		else:
			raise InvalidInteger(read)


class Float(NumberNode):
	def parse(self, text: str) -> ParseResult:
		value, read = utils.get_float(text)
		if value is not None:
			return self._check_in_range_and_return(value, read)
		else:
			raise InvalidFloat(read)

# ------------------
#   Text Arguments
# ------------------


class TextNode(ArgumentNode, ABC):
	def __init__(self, name):
		super().__init__(name)
		self.__min_length = None
		self.__max_length = None

	def at_min_length(self, min_length) -> 'TextNode':
		self.__min_length = min_length
		return self

	def at_max_length(self, max_length) -> 'TextNode':
		self.__max_length = max_length
		return self

	def in_length_range(self, min_length, max_length) -> 'TextNode':
		self.__min_length = min_length
		self.__max_length = max_length
		return self

	def _check_length_in_range_and_return(self, text: str, char_read: int):
		length = len(text)
		if (self.__min_length is not None and length < self.__min_length) or (self.__max_length is not None and length > self.__max_length):
			raise TextLengthOutOfRange(char_read, length, self.__min_length, self.__max_length)
		return ParseResult(text, char_read)


class Text(TextNode):
	"""
	A text argument with no space character
	Just like a single word
	"""
	def parse(self, text: str) -> ParseResult:
		arg = utils.get_element(text)
		return self._check_length_in_range_and_return(arg, len(arg))


class QuotableText(Text):
	QUOTE_CHAR = '"'
	ESCAPE_CHAR = '\\'

	def __init__(self, name):
		super().__init__(name)
		self.empty_allowed = False

	def allow_empty(self):
		self.empty_allowed = True
		return self

	def parse(self, text: str) -> ParseResult:
		if len(text) == 0 or text[0] != self.QUOTE_CHAR:
			return super().parse(text)  # regular text
		collected = []
		i = 1
		escaped = False
		while i < len(text):
			ch = text[i]
			if escaped:
				if ch == self.ESCAPE_CHAR or ch == self.QUOTE_CHAR:
					collected.append(ch)
					escaped = False
				else:
					raise IllegalEscapesUsage(i + 1)
			elif ch == self.ESCAPE_CHAR:
				escaped = True
			elif ch == self.QUOTE_CHAR:
				result = ''.join(collected)
				if not self.empty_allowed and len(result) == 0:
					raise EmptyText(i + 1)
				return self._check_length_in_range_and_return(result, i + 1)
			else:
				collected.append(ch)
			i += 1
		raise UnclosedQuotedString(text)

	# use quote characters to quote suggestions with DIVIDER
	def suggests(self, suggestion: SUGGESTS_CALLBACK) -> 'AbstractNode':
		def quote_wrapper(*args, **kwargs):
			suggestions = []
			for s in suggestion(*args, **kwargs):
				if DIVIDER in s:
					s = json.dumps(s)
				suggestions.append(s)
			return suggestions
		return super().suggests(misc_util.copy_signature(quote_wrapper, suggestion))


class GreedyText(TextNode):
	"""
	A greedy text argument, which will consume all remaining input
	"""
	def parse(self, text: str) -> ParseResult:
		return self._check_length_in_range_and_return(text, len(text))


# -------------------
#   Other Arguments
# -------------------

class Boolean(ArgumentNode):
	"""
	A simple boolean argument, only accepts ``true`` and ``false`` and store them as a bool. Case is ignored
	"""
	def _get_suggestions(self, context: CommandContext) -> Iterable[str]:
		return ['true', 'false']

	def parse(self, text: str) -> ParseResult:
		arg = utils.get_element(text)
		if arg.lower() == 'true':
			value = True
		elif arg.lower() == 'false':
			value = False
		else:
			raise InvalidBoolean(arg)
		return ParseResult(value, len(arg))


class Enumeration(ArgumentNode):
	"""
	A node associating with an Enum class for reading an enum value of the given class
	"""
	def __init__(self, name: str, enum_class: Type[Enum]):
		super().__init__(name)
		self.__enum_class: Type[Enum] = enum_class

	def _get_suggestions(self, context: CommandContext) -> Iterable[str]:
		return map(lambda e: e.name, self.__enum_class)

	def parse(self, text: str) -> ParseResult:
		arg = utils.get_element(text)
		try:
			enum = self.__enum_class[arg]
		except KeyError:
			raise InvalidEnumeration(arg) from None
		else:
			return ParseResult(enum, len(arg))





"""
Simple rcon client implement
"""
import socket
import struct
import time
from logging import Logger
from threading import RLock
from typing import Optional


class PacketType:
	COMMAND_RESPONSE = 0
	COMMAND_REQUEST = 2
	LOGIN_REQUEST = 3
	LOGIN_FAIL = -1
	ENDING_PACKET = 100


class Packet:
	def __init__(self, packet_type=None, payload=None):
		self.packet_id = 0
		self.packet_type = packet_type
		self.payload = payload

	def flush(self):
		data = struct.pack('<ii', self.packet_id, self.packet_type) + bytes(self.payload + '\x00\x00', encoding='utf8')
		return struct.pack('<i', len(data)) + data


class RconConnection:
	BUFFER_SIZE = 2**10

	def __init__(self, address: str, port: int, password: str, *, logger: Optional[Logger] = None):
		self.logger = logger
		self.address = address
		self.port = port
		self.password = password
		self.socket = None
		self.command_lock = RLock()

	def __del__(self):
		self.disconnect()

	def __send(self, data):
		if type(data) is Packet:
			data = data.flush()
		self.socket.send(data)
		time.sleep(0.03)  # MC-72390

	def __receive(self, length):
		data = bytes()
		while len(data) < length:
			data += self.socket.recv(min(self.BUFFER_SIZE, length - len(data)))
		return data

	def __receive_packet(self):
		length = struct.unpack('<i', self.__receive(4))[0]
		data = self.__receive(length)
		packet = Packet()
		packet.packet_id = struct.unpack('<i', data[0:4])[0]
		packet.packet_type = struct.unpack('<i', data[4:8])[0]
		packet.payload = data[8:-2].decode('utf8')
		return packet

	def connect(self) -> bool:
		if self.socket is not None:
			try:
				self.disconnect()
			except:
				pass
		self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.socket.connect((self.address, self.port))
		self.__send(Packet(PacketType.LOGIN_REQUEST, self.password))
		success = self.__receive_packet().packet_id != PacketType.LOGIN_FAIL
		if not success:
			self.disconnect()
		return success

	def disconnect(self):
		if self.socket is None:
			return
		self.socket.close()
		self.socket = None

	def send_command(self, command: str, max_retry_time: int = 3) -> Optional[str]:
		with self.command_lock:
			for i in range(max_retry_time):
				try:
					self.__send(Packet(PacketType.COMMAND_REQUEST, command))
					self.__send(Packet(PacketType.ENDING_PACKET, 'lol'))
					result = ''
					while True:
						packet = self.__receive_packet()
						if packet.payload == 'Unknown request {}'.format(hex(PacketType.ENDING_PACKET)[2:]):
							break
						result += packet.payload
					return result
				except:
					if self.logger is not None:
						self.logger.warning('Rcon Fail to received packet')
					try:
						self.disconnect()
						if self.connect():  # next try
							continue
					except:
						pass
					break
		return None


if __name__ == '__main__':
	rcon = RconConnection('localhost', 25575, 'password')
	print('Login success? ', rcon.connect())
	while True:
		print('Server ->', rcon.send_command(input('Server <- ')))



"""
Some custom exceptions in MCDR that might be useful in plugin dev
"""
from mcdreforged.utils.exception import IllegalCallError, IllegalStateError

__all__ = [
	'IllegalCallError', 'IllegalStateError'
]



"""
Translation support
"""
import collections
import os
from logging import Logger
from typing import Optional, Set

from ruamel.yaml import YAML

from mcdreforged.constants import core_constant
from mcdreforged.minecraft.rtext import RTextBase
from mcdreforged.utils import file_util, translation_util
from mcdreforged.utils.logger import DebugOption
from mcdreforged.utils.types import TranslationStorage, MessageText

LANGUAGE_RESOURCE_DIRECTORY = os.path.join('resources', 'lang')
LANGUAGE_DIRECTORY = os.path.join(core_constant.PACKAGE_PATH, LANGUAGE_RESOURCE_DIRECTORY)


class TranslationManager:
	def __init__(self, logger: Logger):
		self.logger = logger
		self.language = core_constant.DEFAULT_LANGUAGE
		self.translations: TranslationStorage = collections.defaultdict(dict)
		self.available_languages: Set[str] = set()

	def load_translations(self):
		self.translations.clear()
		self.available_languages.clear()
		for file_path in file_util.list_file_with_suffix(LANGUAGE_DIRECTORY, core_constant.LANGUAGE_FILE_SUFFIX):
			language, _ = os.path.basename(file_path).rsplit('.', 1)
			try:
				with open(os.path.join(LANGUAGE_DIRECTORY, file_path), encoding='utf8') as file_handler:
					translations = dict(YAML().load(file_handler))
				for key, text in translations.items():
					self.translations[key][language] = text
				self.available_languages.add(language)
				self.logger.debug('Loaded translation for {} with {} entries'.format(language, len(translations)), option=DebugOption.MCDR)
			except:
				self.logger.exception('Failed to load language {} from "{}"'.format(language, file_path))

	def set_language(self, language: str):
		self.language = language
		if language not in self.available_languages:
			self.logger.warning('Setting language to {} with 0 available translation'.format(language))

	def translate(self, key: str, args: tuple, kwargs: dict, *, allow_failure: bool, language: Optional[str] = None, fallback_language: Optional[str] = None, plugin_translations: Optional[TranslationStorage] = None) -> MessageText:
		if language is None:
			language = self.language
		if plugin_translations is None:
			plugin_translations = {}

		# Translating
		try:
			translated_formatter = translation_util.translate_from_dict(self.translations.get(key, {}), language, fallback_language=fallback_language)
		except KeyError:
			try:
				translated_formatter = translation_util.translate_from_dict(plugin_translations.get(key, {}), language, fallback_language=fallback_language, default=None)
			except KeyError:
				translated_formatter = None

		# Check if there's any rtext inside args and kwargs
		use_rtext = any([isinstance(e, RTextBase) for e in list(args) + list(kwargs.values())])

		# Processing
		if translated_formatter is not None:
			translated_formatter = translated_formatter.strip('\n\r')
			try:
				if use_rtext:
					translated_formatter = RTextBase.format(translated_formatter, *args, **kwargs)
				else:
					translated_formatter = translated_formatter.format(*args, **kwargs)
			except Exception as e:
				raise ValueError('Failed to apply args {} and kwargs {} to translated_text {}: {}'.format(args, kwargs, translated_formatter, e))
			return translated_formatter
		else:
			if not allow_failure:
				raise KeyError('Translation key "{}" not found with language {}, fallback_language {}'.format(key, language, fallback_language))
			self.logger.error('Error translate text "{}" to language {}'.format(key, language))
			return key if not use_rtext else RTextBase.from_any(key)



import os
from typing import TYPE_CHECKING, Callable, List, Tuple, Any

from mcdreforged.command.builder.exception import UnknownArgument
from mcdreforged.command.builder.nodes.arguments import QuotableText
from mcdreforged.command.builder.nodes.basic import Literal
from mcdreforged.command.command_source import CommandSource
from mcdreforged.minecraft.rtext import RText, RAction, RTextList, RStyle, RColor
from mcdreforged.permission.permission_level import PermissionLevel
from mcdreforged.plugin.builtin.mcdreforged_plugin.commands.sub_command import SubCommand
from mcdreforged.plugin.type.plugin import AbstractPlugin
from mcdreforged.plugin.type.regular_plugin import RegularPlugin

if TYPE_CHECKING:
	from mcdreforged.plugin.builtin.mcdreforged_plugin.mcdreforged_plugin import MCDReforgedPlugin


class PluginCommand(SubCommand):
	def __init__(self, mcdr_plugin: 'MCDReforgedPlugin'):
		super().__init__(mcdr_plugin)
		self.plugin_manager = self.mcdr_server.plugin_manager

	def get_command_node(self) -> Literal:
		def plugin_id_node():
			return QuotableText('plugin_id').suggests(lambda: [plg.get_id() for plg in self.plugin_manager.get_regular_plugins()])

		def unloaded_plugin_node():
			return QuotableText('file_name').suggests(lambda: map(os.path.basename, self.server_interface.get_unloaded_plugin_list()))

		def disabled_plugin_node():
			return QuotableText('file_name').suggests(lambda: map(os.path.basename, self.server_interface.get_disabled_plugin_list()))

		return (
			self.control_command_root({'plugin', 'plg'}).
			runs(lambda src: src.reply(self.get_help_message(src, 'mcdr_command.help_message.plugin'))).
			on_error(UnknownArgument, self.on_mcdr_command_unknown_argument).
			then(Literal('list').runs(self.list_plugin)).
			then(Literal('info').then(plugin_id_node().runs(lambda src, ctx: self.show_plugin_info(src, ctx['plugin_id'])))).
			then(Literal('load').then(unloaded_plugin_node().runs(lambda src, ctx: self.load_plugin(src, ctx['file_name'])))).
			then(Literal('enable').then(disabled_plugin_node().runs(lambda src, ctx: self.enable_plugin(src, ctx['file_name'])))).
			then(Literal('reload').then(plugin_id_node().runs(lambda src, ctx: self.reload_plugin(src, ctx['plugin_id'])))).
			then(Literal('unload').then(plugin_id_node().runs(lambda src, ctx: self.unload_plugin(src, ctx['plugin_id'])))).
			then(Literal('disable').then(plugin_id_node().runs(lambda src, ctx: self.disable_plugin(src, ctx['plugin_id'])))).
			then(Literal({'reloadall', 'ra'}).runs(self.reload_all_plugin))
		)

	def list_plugin(self, source: CommandSource):
		not_loaded_plugin_list: List[str] = self.server_interface.get_unloaded_plugin_list()
		disabled_plugin_list: List[str] = self.server_interface.get_disabled_plugin_list()
		current_plugins = list(self.mcdr_server.plugin_manager.get_all_plugins())  # type: List[AbstractPlugin]

		source.reply(self.tr('mcdr_command.list_plugin.info_loaded_plugin', len(current_plugins)))
		for plugin in current_plugins:
			meta = plugin.get_metadata()
			displayed_name = RText(meta.name)
			if not self.can_see_rtext(source):
				displayed_name += RText(' ({})'.format(plugin.get_identifier()), color=RColor.gray)
			texts = RTextList(
				'§7-§r ',
				displayed_name.
				c(RAction.run_command, '{} plugin info {}'.format(self.control_command_prefix, meta.id)).
				h(self.tr('mcdr_command.list_plugin.suggest_info', plugin.get_identifier()))
			)
			if self.can_see_rtext(source) and not plugin.is_permanent():
				texts.append(
					' ',
					RText('[↻]', color=RColor.gray)
					.c(RAction.run_command, '{} plugin reload {}'.format(self.control_command_prefix, meta.id))
					.h(self.tr('mcdr_command.list_plugin.suggest_reload', meta.id)),
					' ',
					RText('[↓]', color=RColor.gray)
					.c(RAction.run_command, '{} plugin unload {}'.format(self.control_command_prefix, meta.id))
					.h(self.tr('mcdr_command.list_plugin.suggest_unload', meta.id)),
					' ',
					RText('[×]', color=RColor.gray)
					.c(RAction.run_command, '{} plugin disable {}'.format(self.control_command_prefix, meta.id))
					.h(self.tr('mcdr_command.list_plugin.suggest_disable', meta.id))
				)
			source.reply(texts)

		def get_file_name(fp) -> Tuple[str, RText]:
			name = os.path.basename(fp)
			name_text = RText(name)
			if source.has_permission(PermissionLevel.PHYSICAL_SERVER_CONTROL_LEVEL):
				name_text.h(fp)
			return name, name_text

		source.reply(self.tr('mcdr_command.list_plugin.info_disabled_plugin', len(disabled_plugin_list)))
		for file_path in disabled_plugin_list:
			file_name, file_name_text = get_file_name(file_path)
			texts = RTextList(RText('- ', color=RColor.gray), file_name_text)
			if self.can_see_rtext(source):
				texts.append(
					' ',
					RText('[✔]', color=RColor.gray)
					.c(RAction.run_command, '{} plugin enable {}'.format(self.control_command_prefix, file_name))
					.h(self.tr('mcdr_command.list_plugin.suggest_enable', file_name))
				)
			source.reply(texts)

		source.reply(self.tr('mcdr_command.list_plugin.info_not_loaded_plugin', len(not_loaded_plugin_list)))
		for file_path in not_loaded_plugin_list:
			file_name, file_name_text = get_file_name(file_path)
			texts = RTextList(RText('- ', color=RColor.gray), file_name_text)
			if self.can_see_rtext(source):
				texts.append(
					' ',
					RText('[↑]', color=RColor.gray)
					.c(RAction.run_command, '{} plugin load {}'.format(self.control_command_prefix, file_name))
					.h(self.tr('mcdr_command.list_plugin.suggest_load', file_name))
				)
			source.reply(texts)

	def show_plugin_info(self, source: CommandSource, plugin_id: str):
		plugin = self.mcdr_server.plugin_manager.get_plugin_from_id(plugin_id)
		if plugin is None:
			source.reply(self.tr('mcdr_command.invalid_plugin_id', plugin_id))
		else:
			meta = plugin.get_metadata()
			source.reply(RTextList(
				RText(meta.name).set_color(RColor.yellow).set_styles(RStyle.bold).h(plugin),
				' ',
				RText('v{}'.format(meta.version), color=RColor.gray)
			))
			source.reply(self.tr('mcdr_command.plugin_info.id', meta.id))
			if meta.author is not None:
				source.reply(self.tr('mcdr_command.plugin_info.author', ', '.join(meta.author)))
			if meta.link is not None:
				source.reply(self.tr('mcdr_command.plugin_info.link', RText(meta.link, color=RColor.blue, styles=RStyle.underlined).c(RAction.open_url, meta.link)))
			if meta.description is not None:
				source.reply(meta.get_description_rtext())

	def __not_loaded_plugin_file_manipulate(self, source: CommandSource, file_name: str, operation_name: str, func: Callable[[str], Any], possible_plugin_path: List[str]):
		plugin_paths = list(filter(lambda fp: fp == file_name, possible_plugin_path))
		if len(plugin_paths) == 0:
			plugin_paths = list(filter(lambda fp: os.path.basename(fp) == file_name, possible_plugin_path))
		if len(plugin_paths) == 0:
			source.reply(self.tr('mcdr_command.invalid_plugin_file_name', file_name))
		else:
			result = self.function_call(source, lambda: func(plugin_paths[0]), operation_name, log_success=False, msg_args=(file_name,))
			if result.return_value is True:
				source.reply(self.tr('mcdr_command.{}.success'.format(operation_name), file_name))
			else:
				source.reply(self.tr('mcdr_command.{}.fail'.format(operation_name), file_name))
			self._print_plugin_operation_result_if_no_error(source, result)

	def __existed_regular_plugin_manipulate(self, source: CommandSource, plugin_id: str, operation_name: str, func: Callable[[RegularPlugin], Any]):
		plugin = self.mcdr_server.plugin_manager.get_regular_plugin_from_id(plugin_id)
		if plugin is None or not plugin.is_regular():
			source.reply(self.tr('mcdr_command.invalid_plugin_id', plugin_id))
		else:
			result = self.function_call(source, lambda: func(plugin), operation_name, log_success=False, msg_args=(plugin.get_name(),))
			if result.return_value is True:
				source.reply(self.tr('mcdr_command.{}.success'.format(operation_name), plugin.get_name()))
			else:
				source.reply(self.tr('mcdr_command.{}.fail'.format(operation_name), plugin.get_name()))
			self._print_plugin_operation_result_if_no_error(source, result)

	def disable_plugin(self, source: CommandSource, plugin_id: str):
		self.__existed_regular_plugin_manipulate(source, plugin_id, 'disable_plugin', lambda plg: self.server_interface.disable_plugin(plg.get_id()))

	def reload_plugin(self, source: CommandSource, plugin_id: str):
		self.__existed_regular_plugin_manipulate(source, plugin_id, 'reload_plugin', lambda plg: self.server_interface.reload_plugin(plg.get_id()))

	def unload_plugin(self, source: CommandSource, plugin_id: str):
		self.__existed_regular_plugin_manipulate(source, plugin_id, 'unload_plugin', lambda plg: self.server_interface.unload_plugin(plg.get_id()))

	def load_plugin(self, source: CommandSource, file_name: str):
		self.__not_loaded_plugin_file_manipulate(source, file_name, 'load_plugin', lambda pth: self.server_interface.load_plugin(pth), self.server_interface.get_unloaded_plugin_list())

	def enable_plugin(self, source: CommandSource, file_name: str):
		self.__not_loaded_plugin_file_manipulate(source, file_name, 'enable_plugin', lambda pth: self.server_interface.enable_plugin(pth), self.server_interface.get_disabled_plugin_list())

	def reload_all_plugin(self, source: CommandSource):
		ret = self.function_call(source, self.mcdr_server.plugin_manager.refresh_all_plugins, 'reload_all_plugin', log_success=False)
		self._print_plugin_operation_result_if_no_error(source, ret)



import collections
from typing import Dict, List, Callable, Any, TYPE_CHECKING, Union

from mcdreforged.command.builder.nodes.basic import Literal
from mcdreforged.minecraft.rtext import RTextBase
from mcdreforged.plugin.plugin_event import EventListener
from mcdreforged.translation.translation_text import RTextMCDRTranslation
from mcdreforged.utils import translation_util
from mcdreforged.utils.types import TranslationStorage, TranslationKeyDictRich, MessageText, \
	TranslationKeyDictNested

if TYPE_CHECKING:
	from mcdreforged.plugin.type.plugin import AbstractPlugin
	from mcdreforged.plugin.plugin_manager import PluginManager

DEFAULT_LISTENER_PRIORITY = 1000


class HelpMessage:
	def __init__(self, plugin: 'AbstractPlugin', prefix: str, message: Union[MessageText, TranslationKeyDictRich], permission: int):
		self.plugin = plugin
		self.prefix = prefix

		self.message: RTextBase
		if isinstance(message, RTextMCDRTranslation):
			self.message = message
		elif isinstance(message, dict):
			self.message = RTextMCDRTranslation.from_translation_dict(message)
		else:
			self.message = RTextBase.from_any(message)

		self.permission = permission
		self.__prefix_lower = self.prefix.lower()

	def __lt__(self, other):
		if not isinstance(other, type(self)):
			return False
		if self.__prefix_lower != other.__prefix_lower:
			return self.__prefix_lower < other.__prefix_lower
		return self.prefix < other.prefix

	def __repr__(self):
		return 'HelpMessage[prefix={},message={},permission={}]'.format(self.prefix, self.message, self.permission)


class PluginCommandNode:
	"""
	A Tuple like data class for tracking the plugin of the node
	"""
	def __init__(self, plugin: 'AbstractPlugin', node: Literal):
		self.plugin = plugin
		self.node = node


class AbstractPluginRegistry:
	def __init__(self):
		self.event_listeners = collections.defaultdict(list)  # type: Dict[str, List[EventListener]]
		self.help_messages: List[HelpMessage] = []
		self.command_roots: List[PluginCommandNode] = []
		self.translations: TranslationStorage = collections.defaultdict(dict)

	def clear(self):
		self.event_listeners.clear()
		self.help_messages.clear()
		self.command_roots.clear()


class PluginRegistry(AbstractPluginRegistry):
	def __init__(self, plugin: 'AbstractPlugin', target_storage: 'PluginRegistryStorage'):
		super().__init__()
		self.plugin = plugin
		self.target_storage = target_storage

	def register_help_message(self, help_message: HelpMessage):
		self.help_messages.append(help_message)

	def register_event_listener(self, event_id: str, listener: EventListener):
		self.event_listeners[event_id].append(listener)

	def register_command(self, node: Literal):
		if not isinstance(node, Literal):
			raise TypeError('Only Literal node is accepted to be a root node')
		self.command_roots.append(PluginCommandNode(self.plugin, node))

	def register_translation(self, language: str, mapping: TranslationKeyDictNested):
		# Translation should be updated immediately
		translation_util.update_storage(self.translations, language, mapping)
		translation_util.update_storage(self.target_storage.translations, language, mapping)


class PluginRegistryStorage(AbstractPluginRegistry):
	def __init__(self, plugin_manager: 'PluginManager'):
		super().__init__()
		self.plugin_manager = plugin_manager

	def collect(self, registry: AbstractPluginRegistry):
		for event_id, plg_listeners in registry.event_listeners.items():
			self.event_listeners[event_id].extend(plg_listeners)
		self.help_messages.extend(registry.help_messages)
		self.command_roots.extend(registry.command_roots)

	def arrange(self):
		self.help_messages.sort()
		for listeners in self.event_listeners.values():
			listeners.sort()

	def export_commands(self, exporter: Callable[[PluginCommandNode], Any]):
		for node in self.command_roots:
			exporter(node)



"""
Analyzing and reacting events related to server
"""

from mcdreforged.info_reactor.abstract_info_reactor import AbstractInfoReactor
from mcdreforged.info_reactor.info import InfoSource
from mcdreforged.info_reactor.server_information import ServerInformation
from mcdreforged.mcdr_state import MCDReforgedFlag
from mcdreforged.plugin.plugin_event import MCDRPluginEvents
from mcdreforged.utils.logger import DebugOption


class ServerReactor(AbstractInfoReactor):
	@property
	def server_info(self) -> ServerInformation:
		return self.mcdr_server.server_information

	def on_server_start(self):
		self.server_info.clear()

	def react(self, info):
		if info.source == InfoSource.SERVER:
			handler = self.mcdr_server.server_handler_manager.get_current_handler()

			if handler.test_server_startup_done(info):
				self.mcdr_server.logger.debug('Server startup detected', option=DebugOption.REACTOR)
				self.mcdr_server.with_flag(MCDReforgedFlag.SERVER_STARTUP)
				self.mcdr_server.plugin_manager.dispatch_event(MCDRPluginEvents.SERVER_STARTUP, ())

			version = handler.parse_server_version(info)
			if version is not None:
				self.mcdr_server.logger.debug('Server version detected: {}'.format(version), option=DebugOption.REACTOR)
				self.server_info.version = version

			ip_and_port = handler.parse_server_address(info)
			if ip_and_port is not None:
				self.mcdr_server.logger.debug('Server ip detected: {}:{}'.format(*ip_and_port), option=DebugOption.REACTOR)
				self.server_info.ip, self.server_info.port = ip_and_port

			if handler.test_rcon_started(info):
				self.mcdr_server.logger.debug('Server rcon started detected', option=DebugOption.REACTOR)
				self.mcdr_server.with_flag(MCDReforgedFlag.SERVER_RCON_READY)
				self.mcdr_server.connect_rcon()

			if handler.test_server_stopping(info):  # notes that it might happen more than once in the server lifecycle
				self.mcdr_server.logger.debug('Server stopping detected', option=DebugOption.REACTOR)
				self.mcdr_server.rcon_manager.disconnect()



import importlib
import json
import os
import sys
from abc import ABC
from typing import IO, Collection

import pkg_resources
from ruamel.yaml import YAML

from mcdreforged.constants import plugin_constant
from mcdreforged.plugin.meta.metadata import Metadata
from mcdreforged.plugin.type.regular_plugin import RegularPlugin
from mcdreforged.utils.exception import BrokenMetadata, IllegalPluginStructure


class MultiFilePlugin(RegularPlugin, ABC):
	def get_fallback_metadata_id(self) -> str:
		raise BrokenMetadata('Missing plugin id in {}'.format(plugin_constant.PLUGIN_META_FILE))

	def open_file(self, file_path: str) -> IO[bytes]:
		raise NotImplementedError()

	def list_directory(self, directory_name: str) -> Collection[str]:
		return os.listdir(os.path.join(self.plugin_path, directory_name))

	def is_own_module(self, module_name: str) -> bool:
		plugin_id = self.get_id()
		return module_name == plugin_id or module_name.startswith('{}.'.format(plugin_id))

	def _get_module_instance(self):
		return importlib.import_module(self.get_metadata().entrypoint)

	def _on_load(self):
		super()._on_load()
		try:
			meta_file = self.open_file(plugin_constant.PLUGIN_META_FILE)
		except:
			raise IllegalPluginStructure('Metadata file {} not found'.format(plugin_constant.PLUGIN_META_FILE)) from None
		with meta_file:
			self._set_metadata(Metadata(json.load(meta_file), plugin=self))
		self.__check_requirements()
		self._check_subdir_legality()

	def _on_unload(self):
		super()._on_unload()
		try:
			sys.path.remove(self.plugin_path)
		except ValueError:
			self.mcdr_server.logger.debug('Fail to remove path "{}" in sys.path for {}'.format(self.plugin_path, self))

	def _on_ready(self):
		sys.path.append(self.plugin_path)
		# It's fail-proof for packed plugin
		try:
			self._load_entry_instance()
		except:
			self.mcdr_server.logger.exception('Fail to load the entry point of {}'.format(self))
		else:
			super()._on_ready()
		self.__register_default_translation()

	def __register_default_translation(self):
		try:
			file_list = self.list_directory(plugin_constant.PLUGIN_TRANSLATION_FILES_PATH)
		except FileNotFoundError:
			return
		for file_name in file_list:
			file_path = os.path.join(plugin_constant.PLUGIN_TRANSLATION_FILES_PATH, file_name)
			try:
				language, file_extension = file_name.rsplit('.', 1)
				with self.open_file(file_path) as file_handler:
					if file_extension == 'json':
						translations = json.load(file_handler)
					elif file_extension == 'yml':
						translations = dict(YAML().load(file_handler))
					else:
						continue
				self.register_translation(language, translations)
			except:
				self.mcdr_server.logger.exception('Fail to load default translation from file {} in {}'.format(file_path, repr(self)))

	def _check_subdir_legality(self):
		"""
		Make sure the only python submodule inside the plugin is named with the plugin id
		:raise IllegalPluginStructure if check failed
		"""
		raise NotImplementedError()

	def __check_requirements(self):
		try:
			req_file = self.open_file(plugin_constant.PLUGIN_REQUIREMENTS_FILE)
		except:
			return
		with req_file:
			requirements = []
			for line in req_file.readlines():
				requirements.append(line.decode('utf8').strip('\r\n'))
		pkg_resources.require(requirements)



import time
from queue import Empty, PriorityQueue
from threading import Lock
from typing import Callable, Any, Optional

from mcdreforged.constants import core_constant
from mcdreforged.executor.thread_executor import ThreadExecutor
from mcdreforged.utils import misc_util
from mcdreforged.utils.logger import DebugOption


class Priority:
	# High priority, executes first
	REGULAR = 0
	INFO = 20
	# Low priority, executes last


class TaskData:
	ID_COUNTER = 0
	ID_COUNTER_LOCK = Lock()

	def __init__(self, func: Callable, priority: int):
		if not isinstance(func, Callable):
			raise TypeError('func should be a callable object')
		self.func = func
		self.priority = priority
		with TaskData.ID_COUNTER_LOCK:
			self.id = TaskData.ID_COUNTER
			TaskData.ID_COUNTER += 1

	def __lt__(self, other):
		if not isinstance(other, type(self)):
			return False
		if self.priority != other.priority:
			return self.priority < other.priority
		else:
			return self.id < other.id


class TaskExecutor(ThreadExecutor):
	def __init__(self, mcdr_server, *, previous_executor: 'Optional[TaskExecutor]' = None):
		super().__init__(mcdr_server)
		self.task_queue = PriorityQueue(maxsize=core_constant.MAX_TASK_QUEUE_SIZE)
		self.__last_tick_time = None
		self.__soft_stopped = False
		if previous_executor is not None:
			self.extract_tasks(previous_executor)

	def extract_tasks(self, other: 'TaskExecutor'):
		count = 0
		while True:
			try:
				task = other.task_queue.get(block=False)
			except Empty:
				break
			else:
				count += 1
				other.task_queue.task_done()  # marked the extracted task as done
				self.task_queue.put(task)
		self.mcdr_server.logger.debug('Extracted {} tasks from the previous executor'.format(count), option=DebugOption.TASK_EXECUTOR)

	def should_keep_looping(self):
		return super().should_keep_looping() and not self.mcdr_server.is_mcdr_exit()

	def soft_stop(self):
		self.__soft_stopped = True

	def enqueue_info_task(self, func: Callable[[], Any], is_user: bool):
		data = TaskData(func, Priority.INFO)
		if is_user:
			self.task_queue.put(data)
		else:
			self.task_queue.put_nowait(data)

	def add_regular_task(self, func: Callable[[], Any], *, block: bool = False, timeout: Optional[float] = None):
		if block:
			func = misc_util.WaitableCallable(func)
		self.task_queue.put(TaskData(func, Priority.REGULAR))
		if block:
			func.wait(timeout)

	def execute_on_thread(self, func: Callable[[], Any], *, block: bool = False, timeout: Optional[float] = None):
		if self.is_on_thread():
			func()
		else:
			self.add_regular_task(func, block=block, timeout=timeout)

	def get_this_tick_time(self) -> float:
		"""
		Return the duration of this tick
		If it's too long, something might go wrong
		"""
		if self.__last_tick_time is None:  # not started yet
			return 0
		return time.monotonic() - self.__last_tick_time

	def tick(self):
		self.__last_tick_time = time.monotonic()
		try:
			task = self.task_queue.get(timeout=0.01)
		except Empty:
			if self.__soft_stopped:
				self.stop()
		else:
			try:
				task.func()
			except:
				self.mcdr_server.logger.exception(self.mcdr_server.tr('task_executor.error'))
			finally:
				self.task_queue.task_done()

	def wait_till_finish_all_task(self):
		if self.is_on_thread():
			raise RuntimeError('Cannot wait in self\'s thread')
		self.task_queue.join()
